"""
ElementsWorker methods for classifications and ML classes.
"""

from peewee import IntegrityError

from arkindex.exceptions import ErrorResponse
from arkindex_worker import logger
from arkindex_worker.cache import CachedClassification, CachedElement
from arkindex_worker.models import Element
from arkindex_worker.utils import (
    DEFAULT_BATCH_SIZE,
    batch_publication,
    make_batches,
    pluralize,
)


class ClassificationMixin:
    def load_corpus_classes(self):
        """
        Load all ML classes available in the worker's corpus and store them in the ``self.classes`` cache.
        """
        corpus_classes = self.api_client.paginate(
            "ListCorpusMLClasses",
            id=self.corpus_id,
        )
        self.classes = {ml_class["name"]: ml_class["id"] for ml_class in corpus_classes}
        logger.info(
            f"Loaded {len(self.classes)} ML {pluralize('class', len(self.classes))} in corpus ({self.corpus_id})"
        )

    def get_ml_class_id(self, ml_class: str) -> str:
        """
        Return the MLClass ID corresponding to the given class name on a specific corpus.

        If no MLClass exists for this class name, a new one is created.
        :param ml_class: Name of the MLClass.
        :returns: ID of the retrieved or created MLClass.
        """
        if not self.classes:
            self.load_corpus_classes()

        ml_class_id = self.classes.get(ml_class)
        if ml_class_id is None:
            logger.info(f"Creating ML class {ml_class} on corpus {self.corpus_id}")
            try:
                response = self.api_client.request(
                    "CreateMLClass", id=self.corpus_id, body={"name": ml_class}
                )
                ml_class_id = self.classes[ml_class] = response["id"]
                logger.debug(f"Created a new ML class {response['id']}")
            except ErrorResponse as e:
                # Only reload for 400 errors
                if e.status_code != 400:
                    raise

                # Reload and make sure we have the class
                logger.info(
                    f"Unable to create the ML class `{ml_class}`. Refreshing corpus classes cache."
                )
                self.load_corpus_classes()
                assert ml_class in self.classes, (
                    f"Missing ML class {ml_class} even after refreshing."
                )
                ml_class_id = self.classes[ml_class]

        return ml_class_id

    def retrieve_ml_class(self, ml_class_id: str) -> str:
        """
        Retrieve the name of the MLClass from its ID.

        :param ml_class_id: ID of the searched MLClass.
        :return: The MLClass's name
        """
        # Load the corpus' MLclasses if they are not available yet
        if not self.classes:
            self.load_corpus_classes()

        # Filter classes by this ml_class_id
        ml_class_name = next(
            filter(
                lambda x: self.classes[x] == ml_class_id,
                self.classes,
            ),
            None,
        )
        assert ml_class_name is not None, (
            f"Missing class with id ({ml_class_id}) in corpus ({self.corpus_id})"
        )
        return ml_class_name

    def create_classification(
        self,
        element: Element | CachedElement,
        ml_class: str,
        confidence: float,
        high_confidence: bool = False,
    ) -> dict[str, str]:
        """
        Create a classification on the given element through the API.

        :param element: The element to create a classification on.
        :param ml_class: Name of the MLClass to use.
        :param confidence: Confidence score for the classification. Must be between 0 and 1.
        :param high_confidence: Whether or not the classification is of high confidence.
        :returns: The created classification, as returned by the ``CreateClassification`` API endpoint.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be an Element or CachedElement"
        )
        assert ml_class and isinstance(ml_class, str), (
            "ml_class shouldn't be null and should be of type str"
        )
        assert isinstance(confidence, float) and 0 <= confidence <= 1, (
            "confidence shouldn't be null and should be a float in [0..1] range"
        )
        assert isinstance(high_confidence, bool), (
            "high_confidence shouldn't be null and should be of type bool"
        )
        if self.is_read_only:
            logger.warning(
                "Cannot create classification as this worker is in read-only mode"
            )
            return
        try:
            created = self.api_client.request(
                "CreateClassification",
                body={
                    "element": str(element.id),
                    "ml_class": self.get_ml_class_id(ml_class),
                    "worker_run_id": self.worker_run_id,
                    "confidence": confidence,
                    "high_confidence": high_confidence,
                },
            )

            if self.use_cache:
                # Store classification in local cache
                try:
                    to_insert = [
                        {
                            "id": created["id"],
                            "element_id": element.id,
                            "class_name": ml_class,
                            "confidence": created["confidence"],
                            "state": created["state"],
                            "worker_run_id": self.worker_run_id,
                        }
                    ]
                    CachedClassification.insert_many(to_insert).execute()
                except IntegrityError as e:
                    logger.warning(
                        f"Couldn't save created classification in local cache: {e}"
                    )
        except ErrorResponse as e:
            # Detect already existing classification
            if e.status_code == 400 and "non_field_errors" in e.content:
                if (
                    "The fields element, worker_run, ml_class must make a unique set."
                    in e.content["non_field_errors"]
                ):
                    logger.warning(
                        f"This worker run has already set {ml_class} on element {element.id}"
                    )
                else:
                    raise
                return

            # Propagate any other API error
            raise

        return created

    @batch_publication
    def create_classifications(
        self,
        element: Element | CachedElement,
        classifications: list[dict[str, str | float | bool]],
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> list[dict[str, str | float | bool]]:
        """
        Create multiple classifications at once on the given element through the API.

        :param element: The element to create classifications on.
        :param classifications: A list of dicts representing a classification each, with the following keys:

            ml_class (str)
                Required. Name of the MLClass to use.
            confidence (float)
                Required. Confidence score for the classification. Must be between 0 and 1.
            high_confidence (bool)
                Optional. Whether or not the classification is of high confidence.

        :param batch_size: The size of each batch, which will be used to split the publication to avoid API errors.

        :returns: List of created classifications, as returned in the ``classifications`` field by
           the ``CreateClassifications`` API endpoint.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be an Element or CachedElement"
        )
        assert classifications and isinstance(classifications, list), (
            "classifications shouldn't be null and should be of type list"
        )

        for index, classification in enumerate(classifications):
            ml_class = classification.get("ml_class")
            assert ml_class and isinstance(ml_class, str), (
                f"Classification at index {index} in classifications: ml_class shouldn't be null and should be of type str"
            )

            confidence = classification.get("confidence")
            assert (
                confidence is not None
                and isinstance(confidence, float)
                and 0 <= confidence <= 1
            ), (
                f"Classification at index {index} in classifications: confidence shouldn't be null and should be a float in [0..1] range"
            )

            high_confidence = classification.get("high_confidence")
            if high_confidence is not None:
                assert isinstance(high_confidence, bool), (
                    f"Classification at index {index} in classifications: high_confidence should be of type bool"
                )

        if self.is_read_only:
            logger.warning(
                "Cannot create classifications as this worker is in read-only mode"
            )
            return

        created_cls = [
            created_cl
            for batch in make_batches(classifications, "classification", batch_size)
            for created_cl in self.api_client.request(
                "CreateClassifications",
                body={
                    "parent": str(element.id),
                    "worker_run_id": self.worker_run_id,
                    "classifications": [
                        {
                            **classification,
                            "ml_class": self.get_ml_class_id(
                                classification["ml_class"]
                            ),
                        }
                        for classification in batch
                    ],
                },
            )["classifications"]
        ]

        for created_cl in created_cls:
            created_cl["class_name"] = self.retrieve_ml_class(created_cl["ml_class"])

        if self.use_cache:
            # Store classifications in local cache
            try:
                to_insert = [
                    {
                        "id": created_cl["id"],
                        "element_id": element.id,
                        "class_name": created_cl.pop("class_name"),
                        "confidence": created_cl["confidence"],
                        "state": created_cl["state"],
                        "worker_run_id": self.worker_run_id,
                    }
                    for created_cl in created_cls
                ]
                CachedClassification.insert_many(to_insert).execute()
            except IntegrityError as e:
                logger.warning(
                    f"Couldn't save created classifications in local cache: {e}"
                )

        return created_cls



import json
import re

import pytest

from arkindex.exceptions import ErrorResponse
from arkindex_worker.models import Element
from arkindex_worker.utils import DEFAULT_BATCH_SIZE
from tests import CORPUS_ID

from . import BASE_API_CALLS


def test_create_sub_element_wrong_element(mock_elements_worker):
    with pytest.raises(
        AssertionError, match="element shouldn't be null and should be of type Element"
    ):
        mock_elements_worker.create_sub_element(
            element=None,
            type="something",
            name="0",
            polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        )

    with pytest.raises(
        AssertionError, match="element shouldn't be null and should be of type Element"
    ):
        mock_elements_worker.create_sub_element(
            element="not element type",
            type="something",
            name="0",
            polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        )


def test_create_sub_element_wrong_type(mock_elements_worker):
    elt = Element({"zone": None})

    with pytest.raises(
        AssertionError, match="type shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_sub_element(
            element=elt,
            type=None,
            name="0",
            polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        )

    with pytest.raises(
        AssertionError, match="type shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_sub_element(
            element=elt,
            type=1234,
            name="0",
            polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        )


def test_create_sub_element_wrong_name(mock_elements_worker):
    elt = Element({"zone": None})

    with pytest.raises(
        AssertionError, match="name shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_sub_element(
            element=elt,
            type="something",
            name=None,
            polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        )

    with pytest.raises(
        AssertionError, match="name shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_sub_element(
            element=elt,
            type="something",
            name=1234,
            polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        )


def test_create_sub_element_wrong_polygon(mock_elements_worker):
    elt = Element({"zone": None})

    with pytest.raises(AssertionError, match="polygon should be None or a list"):
        mock_elements_worker.create_sub_element(
            element=elt,
            type="something",
            name="O",
            polygon="not a polygon",
        )

    with pytest.raises(
        AssertionError, match="polygon should have at least three points"
    ):
        mock_elements_worker.create_sub_element(
            element=elt,
            type="something",
            name="O",
            polygon=[[1, 1], [2, 2]],
        )

    with pytest.raises(
        AssertionError, match="polygon points should be lists of two items"
    ):
        mock_elements_worker.create_sub_element(
            element=elt,
            type="something",
            name="O",
            polygon=[[1, 1, 1], [2, 2, 1], [2, 1, 1], [1, 2, 1]],
        )

    with pytest.raises(
        AssertionError, match="polygon points should be lists of two items"
    ):
        mock_elements_worker.create_sub_element(
            element=elt,
            type="something",
            name="O",
            polygon=[[1], [2], [2], [1]],
        )

    with pytest.raises(
        AssertionError, match="polygon points should be lists of two numbers"
    ):
        mock_elements_worker.create_sub_element(
            element=elt,
            type="something",
            name="O",
            polygon=[["not a coord", 1], [2, 2], [2, 1], [1, 2]],
        )


@pytest.mark.parametrize("confidence", ["lol", "0.2", -1.0, 1.42, float("inf")])
def test_create_sub_element_wrong_confidence(mock_elements_worker, confidence):
    with pytest.raises(
        AssertionError,
        match=re.escape("confidence should be None or a float in [0..1] range"),
    ):
        mock_elements_worker.create_sub_element(
            element=Element({"zone": None}),
            type="something",
            name="blah",
            polygon=[[0, 0], [0, 10], [10, 10], [10, 0], [0, 0]],
            confidence=confidence,
        )


@pytest.mark.parametrize(
    ("image", "error_type", "error_message"),
    [
        (1, AssertionError, "image should be None or string"),
        ("not a uuid", ValueError, "image is not a valid uuid."),
    ],
)
def test_create_sub_element_wrong_image(
    mock_elements_worker, image, error_type, error_message
):
    with pytest.raises(error_type, match=re.escape(error_message)):
        mock_elements_worker.create_sub_element(
            element=Element({"zone": None}),
            type="something",
            name="blah",
            polygon=[[0, 0], [0, 10], [10, 10], [10, 0], [0, 0]],
            image=image,
        )


def test_create_sub_element_wrong_image_and_polygon(mock_elements_worker):
    with pytest.raises(
        AssertionError,
        match=re.escape(
            "An image or a parent with an image is required to create an element with a polygon."
        ),
    ):
        mock_elements_worker.create_sub_element(
            element=Element({"zone": None}),
            type="something",
            name="blah",
            polygon=[[0, 0], [0, 10], [10, 10], [10, 0], [0, 0]],
            image=None,
        )


def test_create_sub_element_api_error(responses, mock_elements_worker):
    elt = Element(
        {
            "id": "12341234-1234-1234-1234-123412341234",
            "corpus": {"id": CORPUS_ID},
            "zone": {"image": {"id": "22222222-2222-2222-2222-222222222222"}},
        }
    )
    responses.add(
        responses.POST,
        "http://testserver/api/v1/elements/create/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_sub_element(
            element=elt,
            type="something",
            name="0",
            polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [("POST", "http://testserver/api/v1/elements/create/")]


@pytest.mark.parametrize("slim_output", [True, False])
def test_create_sub_element(responses, mock_elements_worker, slim_output):
    elt = Element(
        {
            "id": "12341234-1234-1234-1234-123412341234",
            "corpus": {"id": CORPUS_ID},
            "zone": {"image": {"id": "22222222-2222-2222-2222-222222222222"}},
        }
    )
    child_elt = {
        "id": "12345678-1234-1234-1234-123456789123",
        "corpus": {"id": CORPUS_ID},
        "zone": {"image": {"id": "22222222-2222-2222-2222-222222222222"}},
    }
    responses.add(
        responses.POST,
        "http://testserver/api/v1/elements/create/",
        status=200,
        json=child_elt,
    )

    element_creation_response = mock_elements_worker.create_sub_element(
        element=elt,
        type="something",
        name="0",
        polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        slim_output=slim_output,
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "POST",
            "http://testserver/api/v1/elements/create/",
        ),
    ]
    assert json.loads(responses.calls[-1].request.body) == {
        "type": "something",
        "name": "0",
        "image": None,
        "corpus": CORPUS_ID,
        "polygon": [[1, 1], [2, 2], [2, 1], [1, 2]],
        "parent": "12341234-1234-1234-1234-123412341234",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "confidence": None,
    }
    if slim_output:
        assert element_creation_response == "12345678-1234-1234-1234-123456789123"
    else:
        assert Element(element_creation_response) == Element(child_elt)


def test_create_sub_element_confidence(responses, mock_elements_worker):
    elt = Element(
        {
            "id": "12341234-1234-1234-1234-123412341234",
            "corpus": {"id": CORPUS_ID},
            "zone": {"image": {"id": "22222222-2222-2222-2222-222222222222"}},
        }
    )
    responses.add(
        responses.POST,
        "http://testserver/api/v1/elements/create/",
        status=200,
        json={"id": "12345678-1234-1234-1234-123456789123"},
    )

    sub_element_id = mock_elements_worker.create_sub_element(
        element=elt,
        type="something",
        name="0",
        polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        confidence=0.42,
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("POST", "http://testserver/api/v1/elements/create/"),
    ]
    assert json.loads(responses.calls[-1].request.body) == {
        "type": "something",
        "name": "0",
        "image": None,
        "corpus": CORPUS_ID,
        "polygon": [[1, 1], [2, 2], [2, 1], [1, 2]],
        "parent": "12341234-1234-1234-1234-123412341234",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "confidence": 0.42,
    }
    assert sub_element_id == "12345678-1234-1234-1234-123456789123"


@pytest.mark.parametrize(
    ("params", "error_message"),
    [
        (
            {"parent": None, "child": None},
            "parent shouldn't be null and should be of type Element",
        ),
        (
            {"parent": "not an element", "child": None},
            "parent shouldn't be null and should be of type Element",
        ),
        (
            {"parent": Element(zone=None), "child": None},
            "child shouldn't be null and should be of type Element",
        ),
        (
            {"parent": Element(zone=None), "child": "not an element"},
            "child shouldn't be null and should be of type Element",
        ),
    ],
)
def test_create_element_parent_invalid_params(
    mock_elements_worker, params, error_message
):
    with pytest.raises(AssertionError, match=re.escape(error_message)):
        mock_elements_worker.create_element_parent(**params)


def test_create_element_parent_api_error(responses, mock_elements_worker):
    parent = Element({"id": "12341234-1234-1234-1234-123412341234"})
    child = Element({"id": "497f6eca-6276-4993-bfeb-53cbbbba6f08"})
    responses.add(
        responses.POST,
        "http://testserver/api/v1/element/497f6eca-6276-4993-bfeb-53cbbbba6f08/parent/12341234-1234-1234-1234-123412341234/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_element_parent(
            parent=parent,
            child=child,
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "POST",
            "http://testserver/api/v1/element/497f6eca-6276-4993-bfeb-53cbbbba6f08/parent/12341234-1234-1234-1234-123412341234/",
        )
    ]


def test_create_element_parent(responses, mock_elements_worker):
    parent = Element({"id": "12341234-1234-1234-1234-123412341234"})
    child = Element({"id": "497f6eca-6276-4993-bfeb-53cbbbba6f08"})
    responses.add(
        responses.POST,
        "http://testserver/api/v1/element/497f6eca-6276-4993-bfeb-53cbbbba6f08/parent/12341234-1234-1234-1234-123412341234/",
        status=200,
        json={
            "parent": "12341234-1234-1234-1234-123412341234",
            "child": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
        },
    )

    created_element_parent = mock_elements_worker.create_element_parent(
        parent=parent,
        child=child,
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "POST",
            "http://testserver/api/v1/element/497f6eca-6276-4993-bfeb-53cbbbba6f08/parent/12341234-1234-1234-1234-123412341234/",
        ),
    ]
    assert created_element_parent == {
        "parent": "12341234-1234-1234-1234-123412341234",
        "child": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
    }


@pytest.mark.parametrize(
    ("arg_name", "data", "error_message"),
    [
        (
            "parent",
            None,
            "parent shouldn't be null and should be of type Element",
        ),
        (
            "parent",
            "not element type",
            "parent shouldn't be null and should be of type Element",
        ),
        (
            "children",
            None,
            "children shouldn't be null and should be of type list",
        ),
        (
            "children",
            "not a list",
            "children shouldn't be null and should be of type list",
        ),
        (
            "children",
            [
                Element({"id": "11111111-1111-1111-1111-111111111111"}),
                "not element type",
            ],
            "Child at index 1 in children: Should be of type Element",
        ),
    ],
)
def test_create_element_children_wrong_params(
    arg_name, data, error_message, mock_elements_worker
):
    with pytest.raises(AssertionError, match=error_message):
        mock_elements_worker.create_element_children(
            **{
                "parent": Element({"id": "12341234-1234-1234-1234-123412341234"}),
                "children": [
                    Element({"id": "11111111-1111-1111-1111-111111111111"}),
                    Element({"id": "22222222-2222-2222-2222-222222222222"}),
                ],
                # Overwrite with wrong data
                arg_name: data,
            },
        )


def test_create_element_children_api_error(responses, mock_elements_worker):
    parent = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/parent/{parent.id}/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_element_children(
            parent=parent,
            children=[
                Element({"id": "11111111-1111-1111-1111-111111111111"}),
                Element({"id": "22222222-2222-2222-2222-222222222222"}),
            ],
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "POST",
            f"http://testserver/api/v1/element/parent/{parent.id}/",
        )
    ]


@pytest.mark.parametrize("batch_size", [DEFAULT_BATCH_SIZE, 1])
def test_create_element_children(batch_size, responses, mock_elements_worker):
    parent = Element({"id": "12341234-1234-1234-1234-123412341234"})

    first_child = Element({"id": "11111111-1111-1111-1111-111111111111"})
    second_child = Element({"id": "22222222-2222-2222-2222-222222222222"})

    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/parent/{parent.id}/",
        status=200,
        json={"children": []},
    )

    mock_elements_worker.create_element_children(
        parent=parent,
        children=[first_child, second_child],
        batch_size=batch_size,
    )

    bulk_api_calls = [
        (
            "POST",
            f"http://testserver/api/v1/element/parent/{parent.id}/",
        )
    ]
    if batch_size != DEFAULT_BATCH_SIZE:
        bulk_api_calls.append(
            (
                "POST",
                f"http://testserver/api/v1/element/parent/{parent.id}/",
            )
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + len(bulk_api_calls)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + bulk_api_calls

    bodies = []
    first_call_idx = None
    if batch_size > 1:
        first_call_idx = -1
        bodies.append({"children": [first_child.id, second_child.id]})
    else:
        first_call_idx = -2
        bodies.append({"children": [first_child.id]})
        bodies.append({"children": [second_child.id]})

    assert [
        json.loads(bulk_call.request.body)
        for bulk_call in responses.calls[first_call_idx:]
    ] == bodies



import json
import re
from uuid import UUID

import pytest
from playhouse.shortcuts import model_to_dict

from arkindex.exceptions import ErrorResponse
from arkindex_worker.cache import CachedElement, CachedTranscription
from arkindex_worker.models import Element
from arkindex_worker.utils import DEFAULT_BATCH_SIZE
from arkindex_worker.worker.transcription import TextOrientation

from . import BASE_API_CALLS

TRANSCRIPTIONS_SAMPLE = [
    {
        "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
        "confidence": 0.5,
        "text": "The",
    },
    {
        "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
        "confidence": 0.75,
        "text": "first",
        "element_confidence": 0.75,
    },
    {
        "polygon": [[1000, 300], [1200, 300], [1200, 500], [1000, 500]],
        "confidence": 0.9,
        "text": "line",
    },
]


def test_create_element_transcriptions_wrong_element(mock_elements_worker):
    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be an Element or CachedElement",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=None,
            sub_element_type="page",
            transcriptions=TRANSCRIPTIONS_SAMPLE,
        )

    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be an Element or CachedElement",
    ):
        mock_elements_worker.create_element_transcriptions(
            element="not element type",
            sub_element_type="page",
            transcriptions=TRANSCRIPTIONS_SAMPLE,
        )


def test_create_element_transcriptions_wrong_sub_element_type(mock_elements_worker):
    elt = Element({"zone": None})

    with pytest.raises(
        AssertionError,
        match="sub_element_type shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type=None,
            transcriptions=TRANSCRIPTIONS_SAMPLE,
        )

    with pytest.raises(
        AssertionError,
        match="sub_element_type shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type=1234,
            transcriptions=TRANSCRIPTIONS_SAMPLE,
        )


def test_create_element_transcriptions_wrong_transcriptions(mock_elements_worker):
    elt = Element({"zone": None})

    with pytest.raises(
        AssertionError,
        match="transcriptions shouldn't be null and should be of type list",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=None,
        )

    with pytest.raises(
        AssertionError,
        match="transcriptions shouldn't be null and should be of type list",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=1234,
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: text shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "confidence": 0.5,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: text shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "confidence": 0.5,
                    "text": None,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: text shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "confidence": 0.5,
                    "text": 1234,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "text": "word",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "confidence": None,
                    "text": "word",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "confidence": "a wrong confidence",
                    "text": "word",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "confidence": 0,
                    "text": "word",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "confidence": 2.00,
                    "text": "word",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: polygon shouldn't be null and should be of type list",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {"confidence": 0.5, "text": "word"},
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: polygon shouldn't be null and should be of type list",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {"polygon": None, "confidence": 0.5, "text": "word"},
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: polygon shouldn't be null and should be of type list",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {"polygon": "not a polygon", "confidence": 0.5, "text": "word"},
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: polygon should have at least three points",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {"polygon": [[1, 1], [2, 2]], "confidence": 0.5, "text": "word"},
            ],
        )
    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: polygon points should be lists of two items",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[1, 1, 1], [2, 2, 1], [2, 1, 1], [1, 2, 1]],
                    "confidence": 0.5,
                    "text": "word",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: polygon points should be lists of two items",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {"polygon": [[1], [2], [2], [1]], "confidence": 0.5, "text": "word"},
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: polygon points should be lists of two numbers",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [["not a coord", 1], [2, 2], [2, 1], [1, 2]],
                    "confidence": 0.5,
                    "text": "word",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: orientation shouldn't be null and should be of type TextOrientation",
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "confidence": 0.35,
                    "text": "word",
                    "orientation": "uptown",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: element_confidence should be either null or a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=[
                {
                    "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                    "confidence": 0.75,
                    "text": "The",
                },
                {
                    "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                    "confidence": 0.75,
                    "text": "word",
                    "element_confidence": "not a confidence",
                },
            ],
        )


def test_create_element_transcriptions_api_error(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_element_transcriptions(
            element=elt,
            sub_element_type="page",
            transcriptions=TRANSCRIPTIONS_SAMPLE,
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("POST", f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/")
    ]


@pytest.mark.parametrize("batch_size", [DEFAULT_BATCH_SIZE, 2])
def test_create_element_transcriptions(batch_size, responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    if batch_size > 2:
        responses.add(
            responses.POST,
            f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
            status=200,
            json=[
                {
                    "id": "56785678-5678-5678-5678-567856785678",
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "created": True,
                },
                {
                    "id": "67896789-6789-6789-6789-678967896789",
                    "element_id": "22222222-2222-2222-2222-222222222222",
                    "created": False,
                },
                {
                    "id": "78907890-7890-7890-7890-789078907890",
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "created": True,
                },
            ],
        )
    else:
        for transcriptions in [
            [
                ("56785678-5678-5678-5678-567856785678", True),
                ("67896789-6789-6789-6789-678967896789", False),
            ],
            [("78907890-7890-7890-7890-789078907890", True)],
        ]:
            responses.add(
                responses.POST,
                f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
                status=200,
                json=[
                    {
                        "id": tr_id,
                        "element_id": "11111111-1111-1111-1111-111111111111"
                        if created
                        else "22222222-2222-2222-2222-222222222222",
                        "created": created,
                    }
                    for tr_id, created in transcriptions
                ],
            )

    annotations = mock_elements_worker.create_element_transcriptions(
        element=elt,
        sub_element_type="page",
        transcriptions=TRANSCRIPTIONS_SAMPLE,
        batch_size=batch_size,
    )

    bulk_api_calls = [
        (
            "POST",
            f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
        )
    ]
    if batch_size != DEFAULT_BATCH_SIZE:
        bulk_api_calls.append(
            (
                "POST",
                f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
            )
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + len(bulk_api_calls)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + bulk_api_calls

    first_tr = {
        **TRANSCRIPTIONS_SAMPLE[0],
        "orientation": TextOrientation.HorizontalLeftToRight.value,
    }
    second_tr = {
        **TRANSCRIPTIONS_SAMPLE[1],
        "orientation": TextOrientation.HorizontalLeftToRight.value,
    }
    third_tr = {
        **TRANSCRIPTIONS_SAMPLE[2],
        "orientation": TextOrientation.HorizontalLeftToRight.value,
    }
    empty_payload = {
        "element_type": "page",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "transcriptions": [],
        "return_elements": True,
    }

    bodies = []
    first_call_idx = None
    if batch_size > 2:
        first_call_idx = -1
        bodies.append(
            {**empty_payload, "transcriptions": [first_tr, second_tr, third_tr]}
        )
    else:
        first_call_idx = -2
        bodies.append({**empty_payload, "transcriptions": [first_tr, second_tr]})
        bodies.append({**empty_payload, "transcriptions": [third_tr]})

    assert [
        json.loads(bulk_call.request.body)
        for bulk_call in responses.calls[first_call_idx:]
    ] == bodies

    assert annotations == [
        {
            "id": "56785678-5678-5678-5678-567856785678",
            "element_id": "11111111-1111-1111-1111-111111111111",
            "created": True,
        },
        {
            "id": "67896789-6789-6789-6789-678967896789",
            "element_id": "22222222-2222-2222-2222-222222222222",
            "created": False,
        },
        {
            "id": "78907890-7890-7890-7890-789078907890",
            "element_id": "11111111-1111-1111-1111-111111111111",
            "created": True,
        },
    ]


@pytest.mark.parametrize("batch_size", [DEFAULT_BATCH_SIZE, 2])
def test_create_element_transcriptions_with_cache(
    batch_size, responses, mock_elements_worker_with_cache
):
    elt = CachedElement(id="12341234-1234-1234-1234-123412341234", type="thing")

    if batch_size > 2:
        responses.add(
            responses.POST,
            f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
            status=200,
            json=[
                {
                    "id": "56785678-5678-5678-5678-567856785678",
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "created": True,
                },
                {
                    "id": "67896789-6789-6789-6789-678967896789",
                    "element_id": "22222222-2222-2222-2222-222222222222",
                    "created": False,
                },
                {
                    "id": "78907890-7890-7890-7890-789078907890",
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "created": True,
                },
            ],
        )
    else:
        for transcriptions in [
            [
                ("56785678-5678-5678-5678-567856785678", True),
                ("67896789-6789-6789-6789-678967896789", False),
            ],
            [("78907890-7890-7890-7890-789078907890", True)],
        ]:
            responses.add(
                responses.POST,
                f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
                status=200,
                json=[
                    {
                        "id": tr_id,
                        "element_id": "11111111-1111-1111-1111-111111111111"
                        if created
                        else "22222222-2222-2222-2222-222222222222",
                        "created": created,
                    }
                    for tr_id, created in transcriptions
                ],
            )

    annotations = mock_elements_worker_with_cache.create_element_transcriptions(
        element=elt,
        sub_element_type="page",
        transcriptions=TRANSCRIPTIONS_SAMPLE,
        batch_size=batch_size,
    )

    bulk_api_calls = [
        (
            "POST",
            f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
        )
    ]
    if batch_size != DEFAULT_BATCH_SIZE:
        bulk_api_calls.append(
            (
                "POST",
                f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
            )
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + len(bulk_api_calls)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + bulk_api_calls

    first_tr = {
        **TRANSCRIPTIONS_SAMPLE[0],
        "orientation": TextOrientation.HorizontalLeftToRight.value,
    }
    second_tr = {
        **TRANSCRIPTIONS_SAMPLE[1],
        "orientation": TextOrientation.HorizontalLeftToRight.value,
        "element_confidence": 0.75,
    }
    third_tr = {
        **TRANSCRIPTIONS_SAMPLE[2],
        "orientation": TextOrientation.HorizontalLeftToRight.value,
    }
    empty_payload = {
        "element_type": "page",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "transcriptions": [],
        "return_elements": True,
    }

    bodies = []
    first_call_idx = None
    if batch_size > 2:
        first_call_idx = -1
        bodies.append(
            {**empty_payload, "transcriptions": [first_tr, second_tr, third_tr]}
        )
    else:
        first_call_idx = -2
        bodies.append({**empty_payload, "transcriptions": [first_tr, second_tr]})
        bodies.append({**empty_payload, "transcriptions": [third_tr]})

    assert [
        json.loads(bulk_call.request.body)
        for bulk_call in responses.calls[first_call_idx:]
    ] == bodies

    assert annotations == [
        {
            "id": "56785678-5678-5678-5678-567856785678",
            "element_id": "11111111-1111-1111-1111-111111111111",
            "created": True,
        },
        {
            "id": "67896789-6789-6789-6789-678967896789",
            "element_id": "22222222-2222-2222-2222-222222222222",
            "created": False,
        },
        {
            "id": "78907890-7890-7890-7890-789078907890",
            "element_id": "11111111-1111-1111-1111-111111111111",
            "created": True,
        },
    ]

    # Check that created transcriptions and elements were properly stored in SQLite cache
    assert list(CachedElement.select()) == [
        CachedElement(
            id=UUID("11111111-1111-1111-1111-111111111111"),
            parent_id=UUID("12341234-1234-1234-1234-123412341234"),
            type="page",
            polygon="[[100, 150], [700, 150], [700, 200], [100, 200]]",
            worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
        ),
        CachedElement(
            id=UUID("22222222-2222-2222-2222-222222222222"),
            parent_id=UUID("12341234-1234-1234-1234-123412341234"),
            type="page",
            polygon="[[0, 0], [2000, 0], [2000, 3000], [0, 3000]]",
            worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
            confidence=0.75,
        ),
    ]
    assert list(CachedTranscription.select()) == [
        CachedTranscription(
            id=UUID("56785678-5678-5678-5678-567856785678"),
            element_id=UUID("11111111-1111-1111-1111-111111111111"),
            text="The",
            confidence=0.5,
            orientation=TextOrientation.HorizontalLeftToRight.value,
            worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
        ),
        CachedTranscription(
            id=UUID("67896789-6789-6789-6789-678967896789"),
            element_id=UUID("22222222-2222-2222-2222-222222222222"),
            text="first",
            confidence=0.75,
            orientation=TextOrientation.HorizontalLeftToRight.value,
            worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
        ),
        CachedTranscription(
            id=UUID("78907890-7890-7890-7890-789078907890"),
            element_id=UUID("11111111-1111-1111-1111-111111111111"),
            text="line",
            confidence=0.9,
            orientation=TextOrientation.HorizontalLeftToRight.value,
            worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
        ),
    ]


def test_create_element_transcriptions_orientation_with_cache(
    responses, mock_elements_worker_with_cache
):
    elt = CachedElement(id="12341234-1234-1234-1234-123412341234", type="thing")

    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/{elt.id}/transcriptions/bulk/",
        status=200,
        json=[
            {
                "id": "56785678-5678-5678-5678-567856785678",
                "element_id": "11111111-1111-1111-1111-111111111111",
                "created": True,
            },
            {
                "id": "67896789-6789-6789-6789-678967896789",
                "element_id": "22222222-2222-2222-2222-222222222222",
                "created": False,
            },
            {
                "id": "78907890-7890-7890-7890-789078907890",
                "element_id": "11111111-1111-1111-1111-111111111111",
                "created": True,
            },
        ],
    )

    oriented_transcriptions = [
        {
            "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
            "confidence": 0.5,
            "text": "Animula vagula blandula",
        },
        {
            "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
            "confidence": 0.75,
            "text": "Hospes comesque corporis",
            "orientation": TextOrientation.VerticalLeftToRight,
        },
        {
            "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
            "confidence": 0.9,
            "text": "Quae nunc abibis in loca",
            "orientation": TextOrientation.HorizontalRightToLeft,
        },
    ]

    annotations = mock_elements_worker_with_cache.create_element_transcriptions(
        element=elt,
        sub_element_type="page",
        transcriptions=oriented_transcriptions,
    )

    assert json.loads(responses.calls[-1].request.body) == {
        "element_type": "page",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "transcriptions": [
            {
                "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                "confidence": 0.5,
                "text": "Animula vagula blandula",
                "orientation": TextOrientation.HorizontalLeftToRight.value,
            },
            {
                "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                "confidence": 0.75,
                "text": "Hospes comesque corporis",
                "orientation": TextOrientation.VerticalLeftToRight.value,
            },
            {
                "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                "confidence": 0.9,
                "text": "Quae nunc abibis in loca",
                "orientation": TextOrientation.HorizontalRightToLeft.value,
            },
        ],
        "return_elements": True,
    }
    assert annotations == [
        {
            "id": "56785678-5678-5678-5678-567856785678",
            "element_id": "11111111-1111-1111-1111-111111111111",
            "created": True,
        },
        {
            "id": "67896789-6789-6789-6789-678967896789",
            "element_id": "22222222-2222-2222-2222-222222222222",
            "created": False,
        },
        {
            "id": "78907890-7890-7890-7890-789078907890",
            "element_id": "11111111-1111-1111-1111-111111111111",
            "created": True,
        },
    ]

    # Check that the text orientation was properly stored in SQLite cache
    assert list(map(model_to_dict, CachedTranscription.select())) == [
        {
            "id": UUID("56785678-5678-5678-5678-567856785678"),
            "element": {
                "id": UUID("11111111-1111-1111-1111-111111111111"),
                "parent_id": UUID(elt.id),
                "type": "page",
                "image": None,
                "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                "rotation_angle": 0,
                "mirrored": False,
                "initial": False,
                "worker_version_id": None,
                "worker_run_id": UUID("56785678-5678-5678-5678-567856785678"),
                "confidence": None,
            },
            "text": "Animula vagula blandula",
            "confidence": 0.5,
            "orientation": TextOrientation.HorizontalLeftToRight.value,
            "worker_version_id": None,
            "worker_run_id": UUID("56785678-5678-5678-5678-567856785678"),
        },
        {
            "id": UUID("67896789-6789-6789-6789-678967896789"),
            "element": {
                "id": UUID("22222222-2222-2222-2222-222222222222"),
                "parent_id": UUID(elt.id),
                "type": "page",
                "image": None,
                "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000]],
                "rotation_angle": 0,
                "mirrored": False,
                "initial": False,
                "worker_version_id": None,
                "worker_run_id": UUID("56785678-5678-5678-5678-567856785678"),
                "confidence": None,
            },
            "text": "Hospes comesque corporis",
            "confidence": 0.75,
            "orientation": TextOrientation.VerticalLeftToRight.value,
            "worker_version_id": None,
            "worker_run_id": UUID("56785678-5678-5678-5678-567856785678"),
        },
        {
            "id": UUID("78907890-7890-7890-7890-789078907890"),
            "element": {
                "id": UUID("11111111-1111-1111-1111-111111111111"),
                "parent_id": UUID(elt.id),
                "type": "page",
                "image": None,
                "polygon": [[100, 150], [700, 150], [700, 200], [100, 200]],
                "rotation_angle": 0,
                "mirrored": False,
                "initial": False,
                "worker_version_id": None,
                "worker_run_id": UUID("56785678-5678-5678-5678-567856785678"),
                "confidence": None,
            },
            "text": "Quae nunc abibis in loca",
            "confidence": 0.9,
            "orientation": TextOrientation.HorizontalRightToLeft.value,
            "worker_version_id": None,
            "worker_run_id": UUID("56785678-5678-5678-5678-567856785678"),
        },
    ]



import pytest

from tests import PROCESS_ID


@pytest.mark.parametrize(
    ("with_image", "elements"),
    [
        (
            False,
            [
                {
                    "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "type_id": "baaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "name": "element 1",
                    "confidence": 1,
                    "image_id": None,
                    "image_width": None,
                    "image_height": None,
                    "image_url": None,
                    "polygon": None,
                    "rotation_angle": 0,
                    "mirrored": False,
                },
                {
                    "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaa0",
                    "type_id": "baaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "name": "element 2",
                    "confidence": 1,
                    "image_id": None,
                    "image_width": None,
                    "image_height": None,
                    "image_url": None,
                    "polygon": None,
                    "rotation_angle": 0,
                    "mirrored": False,
                },
            ],
        ),
        (
            True,
            [
                {
                    "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "type_id": "baaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "name": "element 1",
                    "confidence": 1,
                    "image_id": "aaa2aaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "image_width": 76,
                    "image_height": 138,
                    "image_url": "http://somewhere.com/iiif/image.jpeg",
                    "polygon": [[0, 0], [0, 40], [20, 40], [20, 0]],
                    "rotation_angle": 0,
                    "mirrored": False,
                },
                {
                    "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaa0",
                    "type_id": "baaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "name": "element 2",
                    "confidence": 1,
                    "image_id": "aaa2aaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "image_width": 138,
                    "image_height": 76,
                    "image_url": "http://somewhere.com/iiif/image.jpeg",
                    "polygon": [[0, 0], [0, 40], [20, 40], [20, 0]],
                    "rotation_angle": 0,
                    "mirrored": False,
                },
            ],
        ),
    ],
)
def test_list_process_elements_with_image(
    responses, mock_elements_worker, with_image, elements
):
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/process/{PROCESS_ID}/elements/?page_size=500&with_count=true&with_image={with_image}",
        status=200,
        json={
            "count": 2,
            "next": None,
            "results": elements,
        },
    )
    assert (
        list(mock_elements_worker.list_process_elements(with_image=with_image))
        == elements
    )



def test_simple_configuration(mock_base_worker_modern_conf, responses):
    # Provide the full configuration directly from the worker run
    responses.add(
        responses.GET,
        "http://testserver/api/v1/workers/runs/56785678-5678-5678-5678-567856785678/configuration/",
        status=200,
        json={"configuration": [{"key": "some_key", "value": "test", "secret": False}]},
    )

    mock_base_worker_modern_conf.configure()

    assert mock_base_worker_modern_conf.config == {"some_key": "test"}
    assert (
        mock_base_worker_modern_conf.user_configuration
        == mock_base_worker_modern_conf.config
    )
    assert mock_base_worker_modern_conf.secrets == {}


def test_empty(mock_base_worker_modern_conf, responses):
    # Provide the full configuration directly from the worker run
    responses.add(
        responses.GET,
        "http://testserver/api/v1/workers/runs/56785678-5678-5678-5678-567856785678/configuration/",
        status=200,
        json={"configuration": []},
    )

    mock_base_worker_modern_conf.configure()

    assert mock_base_worker_modern_conf.config == {}
    assert (
        mock_base_worker_modern_conf.user_configuration
        == mock_base_worker_modern_conf.config
    )
    assert mock_base_worker_modern_conf.secrets == {}


def test_with_secrets(mock_base_worker_modern_conf, responses):
    # Provide the full configuration directly from the worker run
    responses.add(
        responses.GET,
        "http://testserver/api/v1/workers/runs/56785678-5678-5678-5678-567856785678/configuration/",
        status=200,
        json={
            "configuration": [
                {"key": "some_key", "value": "test", "secret": False},
                {
                    "key": "a_secret",
                    "value": "471b9e64-29af-48dc-8bda-1a64a2da0c12",
                    "secret": True,
                },
            ]
        },
    )

    # Provide a secret value
    responses.add(
        responses.GET,
        "http://testserver/api/v1/secret/471b9e64-29af-48dc-8bda-1a64a2da0c12",
        status=200,
        json={
            "id": "471b9e64-29af-48dc-8bda-1a64a2da0c12",
            "name": "a_secret",
            "content": "My super duper secret value",
        },
    )

    mock_base_worker_modern_conf.configure()

    assert mock_base_worker_modern_conf.config == {
        "a_secret": "My super duper secret value",
        "some_key": "test",
    }
    assert (
        mock_base_worker_modern_conf.user_configuration
        == mock_base_worker_modern_conf.config
    )
    assert mock_base_worker_modern_conf.secrets == {
        "a_secret": "My super duper secret value"
    }


def test_with_secrets_ce(mock_base_worker_modern_conf, responses, monkeypatch):
    # Provide the full configuration directly from the worker run
    responses.add(
        responses.GET,
        "http://testserver/api/v1/workers/runs/56785678-5678-5678-5678-567856785678/configuration/",
        status=200,
        json={
            "configuration": [
                {"key": "some_key", "value": "test", "secret": False},
                {
                    "key": "a_secret",
                    "value": "471b9e64-29af-48dc-8bda-1a64a2da0c12",
                    "secret": True,
                },
            ]
        },
    )

    # Remove the RetrieveSecret endpoint to simulate Arkindex CE
    monkeypatch.delitem(
        mock_base_worker_modern_conf.api_client.document.links, "RetrieveSecret"
    )

    mock_base_worker_modern_conf.configure()

    assert mock_base_worker_modern_conf.config == {
        "a_secret": "471b9e64-29af-48dc-8bda-1a64a2da0c12",
        "some_key": "test",
    }
    assert (
        mock_base_worker_modern_conf.user_configuration
        == mock_base_worker_modern_conf.config
    )
    assert mock_base_worker_modern_conf.secrets == {
        # The value is used directly instead of treated as a secret name
        "a_secret": "471b9e64-29af-48dc-8bda-1a64a2da0c12",
    }



"""
Helper methods to download and open IIIF images, and manage polygons.
"""

import base64
import functools
import os
import re
import tempfile
from collections import namedtuple
from collections.abc import Generator, Iterator
from io import BytesIO
from math import ceil
from pathlib import Path
from typing import TYPE_CHECKING

import humanize
import numpy as np
import requests
from PIL import Image
from shapely.affinity import rotate, scale, translate
from shapely.geometry import LinearRing
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_chain,
    wait_fixed,
)

from arkindex_worker import VERSION, logger
from arkindex_worker.utils import pluralize
from teklia_toolbox.requests import should_verify_cert

# Avoid circular imports error when type checking
if TYPE_CHECKING:
    from arkindex_worker.cache import CachedElement
    from arkindex_worker.models import Element

# See http://docs.python-requests.org/en/master/user/advanced/#timeouts
REQUEST_TIMEOUT = (30, 60)

BoundingBox = namedtuple("BoundingBox", ["x", "y", "width", "height"])

# Specific User-Agent to bypass potential server limitations
IIIF_USER_AGENT = f"Teklia/Workers {VERSION}"
# To parse IIIF Urls
IIIF_URL = re.compile(r"\w+:\/{2}.+\/.+\/.+\/.+\/(?P<size>.+)\/!?\d+\/\w+\.\w+")
# Full size of the region
IIIF_FULL = "full"
# Maximum size available
IIIF_MAX = "max"
# Ratios to resize images: 1.0, 0.95, [...], 0.1, 0.05
IMAGE_RATIOS = np.arange(1, 0, -0.05).round(2).tolist()


def update_pillow_image_size_limit(func):
    """
    Update Pillow Image size limit
    """

    @functools.wraps(func)
    def wrapper(
        *args,
        max_image_pixels: str | int | None = os.getenv("ARKINDEX_MAX_IMAGE_PIXELS"),
        **kwargs,
    ):
        """
        Wrapper to update Pillow Image size limit and restore it at the end of the function.

        :param *args: Positional arguments passed to the function.
        :param max_image_pixels: Pillow Image size limit to use.
        :param **kwargs: Keyword arguments passed to the function.
        """
        MAX_IMAGE_PIXELS = Image.MAX_IMAGE_PIXELS

        # Override Pillow Image size limit
        if max_image_pixels is not None:
            max_image_pixels = int(max_image_pixels)
            # Override Pillow limit for detecting decompression bombs, disabled if set to 0
            if max_image_pixels == 0:
                logger.warning(
                    "Pillow Image size limit is completely disabled, make sure you trust the image source."
                )
                Image.MAX_IMAGE_PIXELS = None
            else:
                Image.MAX_IMAGE_PIXELS = max_image_pixels

        try:
            results = func(*args, **kwargs)
        except:
            # Restore initial Pillow Image size limit
            Image.MAX_IMAGE_PIXELS = MAX_IMAGE_PIXELS
            raise

        # Restore initial Pillow Image size limit
        Image.MAX_IMAGE_PIXELS = MAX_IMAGE_PIXELS
        return results

    return wrapper


@update_pillow_image_size_limit
def open_image(
    path: str,
    mode: str | None = "RGB",
    rotation_angle: int | None = 0,
    mirrored: bool | None = False,
) -> Image:
    """
    Open an image from a path or a URL.

    Warns:
    Prefer [arkindex_worker.models.Element.open_image][] whenever possible.

    :param path: Path or URL to open the image from.
       This parameter will be interpreted as a URL when it has a `http` or `https` scheme
       and no file exist with this path locally.
    :param mode: Pillow mode for the image. See [the Pillow documentation](https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes).
    :param rotation_angle: Rotation angle to apply to the image, in degrees.
       If it is not a multiple of 90°, then the rotation can cause empty pixels of
       the mode's default color to be added for padding.
    :param mirrored: Whether or not to mirror the image horizontally.
    :returns: A Pillow image.
    """
    if (
        path.startswith("http://")
        or path.startswith("https://")
        or not Path(path).exists()
    ):
        image = download_image(path)
    else:
        try:
            image = Image.open(path)
        except (OSError, ValueError):
            image = download_image(path)

    if image.mode != mode:
        image = image.convert(mode)

    if mirrored:
        image = image.transpose(Image.Transpose.FLIP_LEFT_RIGHT)

    if rotation_angle:
        image = image.rotate(-rotation_angle, expand=True)

    return image


def download_image(url: str) -> Image:
    """
    Download an image and open it with Pillow.

    :param url: URL of the image.
    :returns: A Pillow image.
    """
    assert url.startswith("http"), "Image URL must be HTTP(S)"

    # Download the image
    # Cannot use stream=True as urllib's responses do not support the seek(int) method,
    # which is explicitly required by Image.open on file-like objects
    try:
        resp = _retried_request(url)
    except requests.exceptions.HTTPError as e:
        if 400 <= e.response.status_code < 500 and (
            # Check if we are trying to download an image
            (parsed_url := IIIF_URL.match(url))
            # Check if we requested the full size of the image
            and parsed_url.group("size") == IIIF_FULL
        ):
            # Limit the size to maximum allowed
            resp = _retried_request(
                url[: parsed_url.start("size")]
                + IIIF_MAX
                + url[parsed_url.end("size") :]
            )
        else:
            raise e

    # Preprocess the image and prepare it for classification
    image = Image.open(BytesIO(resp.content))
    logger.info(
        f"Downloaded image {url} - size={image.size[0]}x{image.size[1]} in {resp.elapsed}"
    )

    return image


def upload_image(image: Image, url: str) -> requests.Response:
    """
    Upload a Pillow image to a URL.

    :param image: Pillow image to upload.
    :param url: Destination URL.
    :returns: The upload response.
    """
    assert url.startswith("http"), "Destination URL for the image must be HTTP(S)"

    # Retrieve a binarized version of the image
    image_bytes = BytesIO()
    image.save(image_bytes, format="jpeg")
    image_bytes.seek(0)

    # Upload the image
    resp = _retried_request(url, method=requests.put, data=image_bytes)
    logger.info(f"Uploaded image to {url} in {resp.elapsed}")

    return resp


def resized_images(
    *args,
    element: "Element",
    max_pixels_short: int | None = None,
    max_pixels_long: int | None = None,
    max_bytes: int | None = None,
    use_base64: bool = False,
    **kwargs,
) -> Iterator[Generator[tempfile._TemporaryFileWrapper | str]]:
    """
    Build resized images according to pixel and byte limits.

    :param *args: Positional arguments passed to [arkindex_worker.models.Element.open_image_tempfile][].
    :param element: Element whose image needs to be resized.
    :param max_pixels_short: Maximum pixel size of the resized images' short side.
    :param max_pixels_long: Maximum pixel size of the resized images' long side.
    :param max_bytes: Maximum byte size of the resized images.
    :param use_base64: Whether or not to encode resized images in base64 before calculating their size.
    :param **kwargs: Keyword arguments passed to [arkindex_worker.models.Element.open_image_tempfile][].
    :returns: An iterator of temporary files for resized images OR an iterator of base64-encoded strings if `use_base64` is set.
    """
    _, _, element_width, element_height = polygon_bounding_box(element.polygon)
    logger.info(
        f"This element's image dimensions are ({element_width} x {element_height})."
    )

    portrait_format = element_width <= element_height
    max_pixels_width, max_pixels_height = (
        (max_pixels_short, max_pixels_long)
        if portrait_format
        else (max_pixels_long, max_pixels_short)
    )

    # The image dimension is already within the pixel limitation, no need to resize the image
    if max_pixels_width and max_pixels_width >= element_width:
        max_pixels_width = None
    if max_pixels_height and max_pixels_height >= element_height:
        max_pixels_height = None

    if (max_pixels_width and element_width > max_pixels_width) or (
        max_pixels_height and element_height > max_pixels_height
    ):
        logger.warning(
            f"Maximum image dimensions supported are ({max_pixels_width or element_width} x {max_pixels_height or element_height})."
        )
        logger.warning("The image will be resized.")

    # No limitations provided, we keep the image initial dimensions
    if max_pixels_width is None and max_pixels_height is None:
        open_image_param, max_value = (
            ("max_height", element_height)
            if portrait_format
            else ("max_width", element_width)
        )
    # A limitation is only given for the height, we resize it
    elif max_pixels_width is None:
        open_image_param, max_value = ("max_height", max_pixels_height)
    # A limitation is only given for the width, we resize it
    elif max_pixels_height is None:
        open_image_param, max_value = ("max_width", max_pixels_width)
    # Limitations are provided for both sides:
    # - we resize only the one with the biggest scale factor
    # - the remaining one will automatically fall within the other limitation
    else:
        width_rescaling_factor = element_width / max_pixels_width
        height_rescaling_factor = element_height / max_pixels_height
        open_image_param, max_value = (
            ("max_height", max_pixels_height)
            if height_rescaling_factor > width_rescaling_factor
            else ("max_width", max_pixels_width)
        )

    resized_pixels = set(
        min(round(ratio * max_value), max_value) for ratio in IMAGE_RATIOS
    )
    for resized_pixel in sorted(resized_pixels, reverse=True):
        with element.open_image_tempfile(
            *args, **{**kwargs, open_image_param: resized_pixel}
        ) as image:
            pillow_image = Image.open(image)
            if (
                pillow_image.width != element_width
                or pillow_image.height != element_height
            ):
                logger.warning(
                    f"The image was resized to ({pillow_image.width} x {pillow_image.height})."
                )

            image_size = Path(image.name).stat().st_size
            if use_base64:
                image = base64.b64encode(Path(image.name).read_bytes()).decode("utf-8")
                image_size = len(image)

            # The image is still too heavy
            if max_bytes and image_size > max_bytes:
                logger.warning(f"The image size is {humanize.naturalsize(image_size)}.")
                logger.warning(
                    f"Maximum image input size supported is {humanize.naturalsize(max_bytes)}."
                )
                logger.warning("The image will be resized.")
                continue

            yield image


def polygon_bounding_box(polygon: list[list[int | float]]) -> BoundingBox:
    """
    Compute the rectangle bounding box of a polygon.

    :param polygon: Polygon to get the bounding box of.
    :returns: Bounding box of this polygon.
    """
    x_coords, y_coords = zip(*polygon, strict=True)
    x, y = min(x_coords), min(y_coords)
    width, height = max(x_coords) - x, max(y_coords) - y
    return BoundingBox(x, y, width, height)


def _retry_log(retry_state, *args, **kwargs):
    logger.warning(
        f"Request to {retry_state.args[0]} failed ({repr(retry_state.outcome.exception())}), "
        f"retrying in {retry_state.idle_for} {pluralize('second', retry_state.idle_for)}"
    )


@retry(
    stop=stop_after_attempt(3),
    # In the event of `requests.RequestException` errors, the call will be retried after 5 seconds, 10 seconds and finally 90 seconds before failing.
    wait=wait_chain(wait_fixed(5), wait_fixed(10), wait_fixed(90)),
    retry=retry_if_exception_type(requests.RequestException),
    before_sleep=_retry_log,
    reraise=True,
)
def _retried_request(url, *args, method=requests.get, **kwargs):
    resp = method(
        url,
        *args,
        headers={"User-Agent": IIIF_USER_AGENT},
        timeout=REQUEST_TIMEOUT,
        verify=should_verify_cert(url),
        **kwargs,
    )
    resp.raise_for_status()
    return resp


def download_tiles(url: str) -> Image:
    """
    Reconstruct a full IIIF image on servers that cannot serve the full-sized image, using tiles.

    :param url: URL of the image.
    :returns: A Pillow image.
    """
    if not url.endswith("/"):
        url += "/"
    logger.debug("Downloading image information")
    info = _retried_request(url + "info.json").json()

    # Use `max` instead of `full` for IIIF 3, since `full` was deprecated in 2.1 then removed in 3.0
    # With IIIF 3, the image's ID will be at `id`, while IIIF 2 will use `@id``
    resize = "max" if "id" in info else "full"

    image_width, image_height = info.get("width"), info.get("height")
    assert image_width and image_height, "Missing image dimensions in info.json"
    assert info.get("tiles"), (
        "Image cannot be retrieved at full size and tiles are not supported"
    )

    # Take the biggest available tile size
    tile = sorted(info["tiles"], key=lambda tile: tile.get("width", 0), reverse=True)[0]
    tile_width = tile["width"]
    # Tile height is optional and defaults to the width
    tile_height = tile.get("height", tile_width)

    full_image = Image.new("RGB", (image_width, image_height))

    for tile_x in range(ceil(image_width / tile_width)):
        for tile_y in range(ceil(image_height / tile_height)):
            region_x = tile_x * tile_width
            region_y = tile_y * tile_height

            # Prevent trying to crop outside the bounds of an image
            region_width = min(tile_width, image_width - region_x)
            region_height = min(tile_height, image_height - region_y)

            logger.debug(f"Downloading tile {tile_x},{tile_y}")
            resp = _retried_request(
                f"{url}{region_x},{region_y},{region_width},{region_height}/{resize}/0/default.jpg"
            )

            tile_img = Image.open(BytesIO(resp.content))

            # Some bad IIIF image server implementations may sometimes return tiles with a few pixels of difference
            # with the expected sizes, causing Pillow to raise ValueError('images do not match').
            actual_width, actual_height = tile_img.size
            if actual_width < region_width or actual_height < region_height:
                # Fail when tiles are too small
                raise ValueError(
                    f"Expected size {region_width}×{region_height} for tile {tile_x},{tile_y}, "
                    f"but got {actual_width}×{actual_height}"
                )

            if actual_width > region_width or actual_height > region_height:
                # Warn and crop when tiles are too large
                logger.warning(
                    f"Cropping tile {tile_x},{tile_y} from {actual_width}×{actual_height} "
                    f"to {region_width}×{region_height}"
                )
                tile_img = tile_img.crop((0, 0, region_width, region_height))

            full_image.paste(
                tile_img,
                box=(
                    region_x,
                    region_y,
                    region_x + region_width,
                    region_y + region_height,
                ),
            )

    return full_image


def trim_polygon(
    polygon: list[list[int]], image_width: int, image_height: int
) -> list[list[int]]:
    """
    Trim a polygon to an image's boundaries, with non-negative coordinates.

    :param polygon: A polygon to trim.
    :param image_width: Width of the image.
    :param image_height: Height of the image.
    :returns: A polygon trimmed to the image's bounds.
    :raises AssertionError: When argument types are invalid or when the trimmed polygon
       is entirely outside of the image's bounds.
    """

    assert isinstance(polygon, list | tuple), (
        "Polygon must be a valid list or tuple of points."
    )
    assert len(polygon) >= 3, "Polygon should have at least three points."
    assert all(isinstance(point, list | tuple) for point in polygon), (
        "Polygon points must be tuples or lists."
    )
    assert all(len(point) == 2 for point in polygon), (
        "Polygon points must be tuples or lists of 2 elements."
    )
    assert all(
        isinstance(point[0], int) and isinstance(point[1], int) for point in polygon
    ), "Polygon point coordinates must be integers."
    assert any(
        point[0] <= image_width and point[1] <= image_height for point in polygon
    ), "This polygon is entirely outside the image's bounds."

    return [
        [
            min(image_width, max(0, x)),
            min(image_height, max(0, y)),
        ]
        for x, y in polygon
    ]


def revert_orientation(
    element: "Element | CachedElement",
    polygon: list[list[int | float]],
    reverse: bool = False,
) -> list[list[int]]:
    """
    Update the coordinates of the polygon of a child element based on the orientation of
    its parent.

    This method should be called before sending any polygon to Arkindex, to undo the possible
    orientation applied by [arkindex_worker.models.Element.open_image][].

    In some cases, we want to apply the parent's orientation on the child's polygon instead. This is done
    by enabling `reverse=True`.

    :param element: Parent element.
    :param polygon: Polygon corresponding to the child element.
    :param reverse: Whether we should revert or apply the parent's orientation.
    :return: A polygon with updated coordinates.
    """
    from arkindex_worker.cache import CachedElement
    from arkindex_worker.models import Element

    assert element and isinstance(element, Element | CachedElement), (
        "element shouldn't be null and should be an Element or CachedElement"
    )
    assert polygon and isinstance(polygon, list), (
        "polygon shouldn't be null and should be a list"
    )
    assert isinstance(reverse, bool), "reverse should be a bool"
    # Rotating with Pillow can cause it to move the image around, as the image cannot have negative coordinates
    # and must be a rectangle.  This means the origin point of any coordinates from an image is invalid, and the
    # center of the bounding box of the rotated image is different from the center of the element's bounding box.
    # To properly undo the mirroring and rotation implicitly applied by open_image, we first need to find the center
    # of the rotated bounding box.
    if isinstance(element, Element):
        assert element.zone and element.zone.polygon, (
            "element should have a zone and a polygon"
        )
        parent_ring = LinearRing(element.zone.polygon)
    elif isinstance(element, CachedElement):
        assert element.polygon, "cached element should have a polygon"
        parent_ring = LinearRing(element.polygon)

    rotated_ring = rotate(parent_ring, element.rotation_angle, origin="center")

    # This rotated ring might have negative coordinates, so we get the vector that Pillow applies to offset the
    # image to non-negative coordinates using the rotated bounding box.
    offset_x, offset_y, _, _ = rotated_ring.bounds

    # This uses the same calculation as what Shapely does for rotate/scale(origin='center').
    # We will use this below to rotate around the center of the parent bounding box and not of each child polygon.
    # https://github.com/Toblerity/Shapely/blob/462de3aa7a8bbd80408762a2d5aaf84b04476e4d/shapely/affinity.py#L98-L101
    minx, miny, maxx, maxy = parent_ring.bounds
    origin = ((maxx + minx) / 2.0, (maxy + miny) / 2.0)

    ring = LinearRing(polygon)

    if reverse:
        # Apply the parent's orientation on the child's polygon
        # Apply mirroring
        if element.mirrored:
            ring = scale(ring, xfact=-1, origin=origin)
        # Apply rotation
        if element.rotation_angle:
            ring = rotate(ring, element.rotation_angle, origin=origin)
        # At last translate coordinates offset
        ring = translate(ring, xoff=-offset_x, yoff=-offset_y)
    else:
        # First undo the negative coordinates offset, since this is the last step of the original transformation
        ring = translate(ring, xoff=offset_x, yoff=offset_y)
        # Revert any rotation
        if element.rotation_angle:
            ring = rotate(ring, -element.rotation_angle, origin=origin)
        # Revert any mirroring
        if element.mirrored:
            ring = scale(ring, xfact=-1, origin=origin)

    return [[int(x), int(y)] for x, y in ring.coords]



"""
ElementsWorker methods for transcriptions.
"""

from collections.abc import Iterable
from enum import Enum
from warnings import warn

from peewee import IntegrityError

from arkindex_worker import logger
from arkindex_worker.cache import CachedElement, CachedTranscription
from arkindex_worker.models import Element
from arkindex_worker.utils import DEFAULT_BATCH_SIZE, batch_publication, make_batches


class TextOrientation(Enum):
    """
    Orientation of a transcription's text.
    """

    HorizontalLeftToRight = "horizontal-lr"
    """
    The text is read from top to bottom then left to right.
    This is the default when no orientation is specified.
    """

    HorizontalRightToLeft = "horizontal-rl"
    """
    The text is read from top to bottom then right to left.
    """

    VerticalRightToLeft = "vertical-rl"
    """
    The text is read from right to left then top to bottom.
    """

    VerticalLeftToRight = "vertical-lr"
    """
    The text is read from left to right then top to bottom.
    """


class TranscriptionMixin:
    def create_transcription(
        self,
        element: Element | CachedElement,
        text: str,
        confidence: float,
        orientation: TextOrientation = TextOrientation.HorizontalLeftToRight,
    ) -> dict[str, str | float] | None:
        """
        Create a transcription on the given element through the API.

        :param element: Element to create a transcription on.
        :param text: Text of the transcription.
        :param confidence: Confidence score, between 0 and 1.
        :param orientation: Orientation of the transcription's text.
        :returns: A dict as returned by the ``CreateTranscription`` API endpoint,
           or None if the worker is in read-only mode.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be an Element or CachedElement"
        )
        assert text and isinstance(text, str), (
            "text shouldn't be null and should be of type str"
        )
        assert orientation and isinstance(orientation, TextOrientation), (
            "orientation shouldn't be null and should be of type TextOrientation"
        )
        assert isinstance(confidence, float) and 0 <= confidence <= 1, (
            "confidence shouldn't be null and should be a float in [0..1] range"
        )

        if self.is_read_only:
            logger.warning(
                "Cannot create transcription as this worker is in read-only mode"
            )
            return

        created = self.api_client.request(
            "CreateTranscription",
            id=element.id,
            body={
                "text": text,
                "worker_run_id": self.worker_run_id,
                "confidence": confidence,
                "orientation": orientation.value,
            },
        )

        if self.use_cache:
            # Store transcription in local cache
            try:
                to_insert = [
                    {
                        "id": created["id"],
                        "element_id": element.id,
                        "text": created["text"],
                        "confidence": created["confidence"],
                        "orientation": created["orientation"],
                        "worker_run_id": self.worker_run_id,
                    }
                ]
                CachedTranscription.insert_many(to_insert).execute()
            except IntegrityError as e:
                logger.warning(
                    f"Couldn't save created transcription in local cache: {e}"
                )

        return created

    @batch_publication
    def create_transcriptions(
        self,
        transcriptions: list[dict[str, str | float | TextOrientation | None]],
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> list[dict[str, str | float]]:
        """
        Create multiple transcriptions at once on existing elements through the API,
        and creates [CachedTranscription][arkindex_worker.cache.CachedTranscription] instances if cache support is enabled.

        :param transcriptions: A list of dicts representing a transcription each, with the following keys:

            element_id (str)
                Required. UUID of the element to create this transcription on.
            text (str)
                Required. Text of the transcription.
            confidence (float)
                Required. Confidence score between 0 and 1.
            orientation (TextOrientation)
                Optional. Orientation of the transcription's text.

        :param batch_size: The size of each batch, which will be used to split the publication to avoid API errors.

        :returns: A list of dicts as returned in the ``transcriptions`` field by the ``CreateTranscriptions`` API endpoint.
        """

        assert transcriptions and isinstance(transcriptions, list), (
            "transcriptions shouldn't be null and should be of type list"
        )

        # Create shallow copies of every transcription to avoid mutating the original payload
        transcriptions_payload = list(map(dict, transcriptions))

        for index, transcription in enumerate(transcriptions_payload):
            element_id = transcription.get("element_id")
            assert element_id and isinstance(element_id, str), (
                f"Transcription at index {index} in transcriptions: element_id shouldn't be null and should be of type str"
            )

            text = transcription.get("text")
            assert text and isinstance(text, str), (
                f"Transcription at index {index} in transcriptions: text shouldn't be null and should be of type str"
            )

            confidence = transcription.get("confidence")
            assert (
                confidence is not None
                and isinstance(confidence, float)
                and 0 <= confidence <= 1
            ), (
                f"Transcription at index {index} in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
            )

            orientation = transcription.get(
                "orientation", TextOrientation.HorizontalLeftToRight
            )
            assert orientation and isinstance(orientation, TextOrientation), (
                f"Transcription at index {index} in transcriptions: orientation shouldn't be null and should be of type TextOrientation"
            )
            if orientation:
                transcription["orientation"] = orientation.value

        if self.is_read_only:
            logger.warning(
                "Cannot create transcription as this worker is in read-only mode"
            )
            return

        created_trs = [
            created_tr
            for batch in make_batches(
                transcriptions_payload, "transcription", batch_size
            )
            for created_tr in self.api_client.request(
                "CreateTranscriptions",
                body={
                    "worker_run_id": self.worker_run_id,
                    "transcriptions": batch,
                },
            )["transcriptions"]
        ]

        if self.use_cache:
            # Store transcriptions in local cache
            try:
                to_insert = [
                    {
                        "id": created_tr["id"],
                        "element_id": created_tr["element_id"],
                        "text": created_tr["text"],
                        "confidence": created_tr["confidence"],
                        "orientation": created_tr["orientation"],
                        "worker_run_id": self.worker_run_id,
                    }
                    for created_tr in created_trs
                ]
                CachedTranscription.insert_many(to_insert).execute()
            except IntegrityError as e:
                logger.warning(
                    f"Couldn't save created transcriptions in local cache: {e}"
                )

        return created_trs

    @batch_publication
    def create_element_transcriptions(
        self,
        element: Element | CachedElement,
        sub_element_type: str,
        transcriptions: list[dict[str, str | float]],
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> dict[str, str | bool]:
        """
        Create multiple elements and transcriptions at once on a single parent element through the API.

        :param element: Element to create elements and transcriptions on.
        :param sub_element_type: Slug of the element type to use for the new elements.
        :param transcriptions: A list of dicts representing an element and transcription each, with the following keys:

            polygon (list(list(int or float)))
                Required. Polygon of the element.
            text (str)
                Required. Text of the transcription.
            confidence (float)
                Required. Confidence score between 0 and 1.
            orientation ([TextOrientation][arkindex_worker.worker.transcription.TextOrientation])
                Optional. Orientation of the transcription's text.
            element_confidence (float)
                Optional. Confidence score of the element between 0 and 1.

        :param batch_size: The size of each batch, which will be used to split the publication to avoid API errors.

        :returns: A list of dicts as returned by the ``CreateElementTranscriptions`` API endpoint.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be an Element or CachedElement"
        )
        assert sub_element_type and isinstance(sub_element_type, str), (
            "sub_element_type shouldn't be null and should be of type str"
        )
        assert transcriptions and isinstance(transcriptions, list), (
            "transcriptions shouldn't be null and should be of type list"
        )

        # Create shallow copies of every transcription to avoid mutating the original payload
        transcriptions_payload = list(map(dict, transcriptions))

        for index, transcription in enumerate(transcriptions_payload):
            text = transcription.get("text")
            assert text and isinstance(text, str), (
                f"Transcription at index {index} in transcriptions: text shouldn't be null and should be of type str"
            )

            confidence = transcription.get("confidence")
            assert (
                confidence is not None
                and isinstance(confidence, float)
                and 0 <= confidence <= 1
            ), (
                f"Transcription at index {index} in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
            )

            orientation = transcription.get(
                "orientation", TextOrientation.HorizontalLeftToRight
            )
            assert orientation and isinstance(orientation, TextOrientation), (
                f"Transcription at index {index} in transcriptions: orientation shouldn't be null and should be of type TextOrientation"
            )
            if orientation:
                transcription["orientation"] = orientation.value

            polygon = transcription.get("polygon")
            assert polygon and isinstance(polygon, list), (
                f"Transcription at index {index} in transcriptions: polygon shouldn't be null and should be of type list"
            )
            assert len(polygon) >= 3, (
                f"Transcription at index {index} in transcriptions: polygon should have at least three points"
            )
            assert all(
                isinstance(point, list) and len(point) == 2 for point in polygon
            ), (
                f"Transcription at index {index} in transcriptions: polygon points should be lists of two items"
            )
            assert all(
                isinstance(coord, int | float) for point in polygon for coord in point
            ), (
                f"Transcription at index {index} in transcriptions: polygon points should be lists of two numbers"
            )

            element_confidence = transcription.get("element_confidence")
            assert element_confidence is None or (
                isinstance(element_confidence, float) and 0 <= element_confidence <= 1
            ), (
                f"Transcription at index {index} in transcriptions: element_confidence should be either null or a float in [0..1] range"
            )

        if self.is_read_only:
            logger.warning(
                "Cannot create transcriptions as this worker is in read-only mode"
            )
            return

        annotations = [
            annotation
            for batch in make_batches(
                transcriptions_payload, "transcription", batch_size
            )
            for annotation in self.api_client.request(
                "CreateElementTranscriptions",
                id=element.id,
                body={
                    "element_type": sub_element_type,
                    "worker_run_id": self.worker_run_id,
                    "transcriptions": batch,
                    "return_elements": True,
                },
            )
        ]

        for annotation in annotations:
            if annotation["created"]:
                logger.debug(
                    f"A sub_element of {element.id} with type {sub_element_type} was created during transcriptions bulk creation"
                )

        if self.use_cache:
            # Store transcriptions and their associated element (if created) in local cache
            created_ids = set()
            elements_to_insert = []
            transcriptions_to_insert = []
            for index, annotation in enumerate(annotations):
                transcription = transcriptions[index]

                if annotation["element_id"] not in created_ids:
                    # Even if the API says the element already existed in the DB,
                    # we need to check if it is available in the local cache.
                    # Peewee does not have support for SQLite's INSERT OR IGNORE,
                    # so we do the check here, element by element.
                    try:
                        CachedElement.get_by_id(annotation["element_id"])
                    except CachedElement.DoesNotExist:
                        elements_to_insert.append(
                            {
                                "id": annotation["element_id"],
                                "parent_id": element.id,
                                "type": sub_element_type,
                                "image_id": element.image_id,
                                "polygon": transcription["polygon"],
                                "worker_run_id": self.worker_run_id,
                                "confidence": transcription.get("element_confidence"),
                            }
                        )

                    created_ids.add(annotation["element_id"])

                transcriptions_to_insert.append(
                    {
                        "id": annotation["id"],
                        "element_id": annotation["element_id"],
                        "text": transcription["text"],
                        "confidence": transcription["confidence"],
                        "orientation": transcription.get(
                            "orientation", TextOrientation.HorizontalLeftToRight
                        ).value,
                        "worker_run_id": self.worker_run_id,
                    }
                )

            try:
                CachedElement.insert_many(elements_to_insert).execute()
                CachedTranscription.insert_many(transcriptions_to_insert).execute()
            except IntegrityError as e:
                logger.warning(
                    f"Couldn't save created transcriptions in local cache: {e}"
                )

        return annotations

    def list_transcriptions(
        self,
        element: Element | CachedElement,
        element_type: str | None = None,
        recursive: bool | None = None,
        worker_version: str | bool | None = None,
        worker_run: str | bool | None = None,
    ) -> Iterable[dict] | Iterable[CachedTranscription]:
        """
        List transcriptions on an element.

        Warns:
        ----
        The following parameters are **deprecated**:

        - `worker_version` in favor of `worker_run`

        :param element: The element to list transcriptions on.
        :param element_type: Restrict to transcriptions whose elements have an element type with this slug.
        :param recursive: Include transcriptions of any descendant of this element, recursively.
        :param worker_version: **Deprecated** Restrict to transcriptions created by a worker version with this UUID. Set to False to look for manually created transcriptions.
        :param worker_run: Restrict to transcriptions created by a worker run with this UUID. Set to False to look for manually created transcriptions.
        :returns: An iterable of dicts representing each transcription,
           or an iterable of CachedTranscription when cache support is enabled.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be an Element or CachedElement"
        )
        query_params = {}
        if element_type:
            assert isinstance(element_type, str), "element_type should be of type str"
            query_params["element_type"] = element_type
        if recursive is not None:
            assert isinstance(recursive, bool), "recursive should be of type bool"
            query_params["recursive"] = recursive
        if worker_version is not None:
            warn(
                "`worker_version` usage is deprecated. Consider using `worker_run` instead.",
                DeprecationWarning,
                stacklevel=1,
            )
            assert isinstance(worker_version, str | bool), (
                "worker_version should be of type str or bool"
            )
            if isinstance(worker_version, bool):
                assert worker_version is False, (
                    "if of type bool, worker_version can only be set to False"
                )
            query_params["worker_version"] = worker_version
        if worker_run is not None:
            assert isinstance(worker_run, str | bool), (
                "worker_run should be of type str or bool"
            )
            if isinstance(worker_run, bool):
                assert worker_run is False, (
                    "if of type bool, worker_run can only be set to False"
                )
            query_params["worker_run"] = worker_run

        if not self.use_cache:
            return self.api_client.paginate(
                "ListTranscriptions", id=element.id, **query_params
            )

        if not recursive:
            # In this case we don't have to return anything, it's easier to use an
            # impossible condition (False) rather than filtering by type for nothing
            if element_type and element_type != element.type:
                return CachedTranscription.select().where(False)
            transcriptions = CachedTranscription.select().where(
                CachedTranscription.element_id == element.id
            )
        else:
            base_case = (
                CachedElement.select()
                .where(CachedElement.id == element.id)
                .cte("base", recursive=True)
            )
            recursive = CachedElement.select().join(
                base_case, on=(CachedElement.parent_id == base_case.c.id)
            )
            cte = base_case.union_all(recursive)
            transcriptions = (
                CachedTranscription.select()
                .join(cte, on=(CachedTranscription.element_id == cte.c.id))
                .with_cte(cte)
            )

            if element_type:
                transcriptions = transcriptions.where(cte.c.type == element_type)

        if worker_version is not None:
            # If worker_version=False, filter by manual worker_version e.g. None
            worker_version_id = worker_version or None
            if worker_version_id:
                transcriptions = transcriptions.where(
                    CachedTranscription.worker_version_id == worker_version_id
                )
            else:
                transcriptions = transcriptions.where(
                    CachedTranscription.worker_version_id.is_null()
                )

        if worker_run is not None:
            # If worker_run=False, filter by manual worker_run e.g. None
            worker_run_id = worker_run or None
            if worker_run_id:
                transcriptions = transcriptions.where(
                    CachedTranscription.worker_run_id == worker_run_id
                )
            else:
                transcriptions = transcriptions.where(
                    CachedTranscription.worker_run_id.is_null()
                )

        return transcriptions



from uuid import UUID

import pytest

from arkindex_worker.cache import CachedElement, CachedTranscription
from arkindex_worker.models import Element

from . import BASE_API_CALLS


def test_list_transcriptions_wrong_element(mock_elements_worker):
    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be an Element or CachedElement",
    ):
        mock_elements_worker.list_transcriptions(element=None)

    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be an Element or CachedElement",
    ):
        mock_elements_worker.list_transcriptions(element="not element type")


def test_list_transcriptions_wrong_element_type(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="element_type should be of type str"):
        mock_elements_worker.list_transcriptions(
            element=elt,
            element_type=1234,
        )


def test_list_transcriptions_wrong_recursive(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="recursive should be of type bool"):
        mock_elements_worker.list_transcriptions(
            element=elt,
            recursive="not bool",
        )


def test_list_transcriptions_wrong_worker_run(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError, match="worker_run should be of type str or bool"
    ):
        mock_elements_worker.list_transcriptions(
            element=elt,
            worker_run=1234,
        )


def test_list_transcriptions_wrong_worker_version(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    # WARNING: pytest.deprecated_call must be placed BEFORE pytest.raises, otherwise `match` argument won't be checked
    with (
        pytest.deprecated_call(
            match="`worker_version` usage is deprecated. Consider using `worker_run` instead."
        ),
        pytest.raises(
            AssertionError, match="worker_version should be of type str or bool"
        ),
    ):
        mock_elements_worker.list_transcriptions(
            element=elt,
            worker_version=1234,
        )


def test_list_transcriptions_wrong_bool_worker_run(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError, match="if of type bool, worker_run can only be set to False"
    ):
        mock_elements_worker.list_transcriptions(
            element=elt,
            worker_run=True,
        )


def test_list_transcriptions_wrong_bool_worker_version(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    # WARNING: pytest.deprecated_call must be placed BEFORE pytest.raises, otherwise `match` argument won't be checked
    with (
        pytest.deprecated_call(
            match="`worker_version` usage is deprecated. Consider using `worker_run` instead."
        ),
        pytest.raises(
            AssertionError,
            match="if of type bool, worker_version can only be set to False",
        ),
    ):
        mock_elements_worker.list_transcriptions(
            element=elt,
            worker_version=True,
        )


def test_list_transcriptions_api_error(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.GET,
        "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/",
        status=418,
    )

    with pytest.raises(
        Exception, match="Stopping pagination as data will be incomplete"
    ):
        next(mock_elements_worker.list_transcriptions(element=elt))

    assert len(responses.calls) == len(BASE_API_CALLS) + 5
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        # We do 5 retries
        (
            "GET",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/",
        ),
        (
            "GET",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/",
        ),
        (
            "GET",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/",
        ),
        (
            "GET",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/",
        ),
        (
            "GET",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/",
        ),
    ]


def test_list_transcriptions(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    trans = [
        {
            "id": "0000",
            "text": "hey",
            "confidence": 0.42,
            "worker_version_id": "56785678-5678-5678-5678-567856785678",
            "worker_run_id": "56785678-5678-5678-5678-567856785678",
            "element": None,
        },
        {
            "id": "1111",
            "text": "it's",
            "confidence": 0.42,
            "worker_version_id": "56785678-5678-5678-5678-567856785678",
            "worker_run_id": "56785678-5678-5678-5678-567856785678",
            "element": None,
        },
        {
            "id": "2222",
            "text": "me",
            "confidence": 0.42,
            "worker_version_id": "56785678-5678-5678-5678-567856785678",
            "worker_run_id": "56785678-5678-5678-5678-567856785678",
            "element": None,
        },
    ]
    responses.add(
        responses.GET,
        "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/",
        status=200,
        json={
            "count": 3,
            "next": None,
            "results": trans,
        },
    )

    for idx, transcription in enumerate(
        mock_elements_worker.list_transcriptions(element=elt)
    ):
        assert transcription == trans[idx]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "GET",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/",
        ),
    ]


def test_list_transcriptions_manual_worker_version(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    trans = [
        {
            "id": "0000",
            "text": "hey",
            "confidence": 0.42,
            "worker_version_id": None,
            "worker_run_id": None,
            "element": None,
        }
    ]
    responses.add(
        responses.GET,
        "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/?worker_version=False",
        status=200,
        json={
            "count": 1,
            "next": None,
            "results": trans,
        },
    )

    with pytest.deprecated_call(
        match="`worker_version` usage is deprecated. Consider using `worker_run` instead."
    ):
        for idx, transcription in enumerate(
            mock_elements_worker.list_transcriptions(element=elt, worker_version=False)
        ):
            assert transcription == trans[idx]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "GET",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/?worker_version=False",
        ),
    ]


def test_list_transcriptions_manual_worker_run(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    trans = [
        {
            "id": "0000",
            "text": "hey",
            "confidence": 0.42,
            "worker_version_id": None,
            "worker_run_id": None,
            "element": None,
        }
    ]
    responses.add(
        responses.GET,
        "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/?worker_run=False",
        status=200,
        json={
            "count": 1,
            "next": None,
            "results": trans,
        },
    )

    for idx, transcription in enumerate(
        mock_elements_worker.list_transcriptions(element=elt, worker_run=False)
    ):
        assert transcription == trans[idx]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "GET",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/transcriptions/?worker_run=False",
        ),
    ]


@pytest.mark.usefixtures("_mock_cached_transcriptions")
@pytest.mark.parametrize(
    ("filters", "expected_ids"),
    [
        # Filter on element should give first and sixth transcription
        (
            {
                "element": CachedElement(
                    id="11111111-1111-1111-1111-111111111111", type="page"
                ),
            },
            (
                "11111111-1111-1111-1111-111111111111",
                "66666666-6666-6666-6666-666666666666",
            ),
        ),
        # Filter on element and element_type should give first and sixth transcription
        (
            {
                "element": CachedElement(
                    id="11111111-1111-1111-1111-111111111111", type="page"
                ),
                "element_type": "page",
            },
            (
                "11111111-1111-1111-1111-111111111111",
                "66666666-6666-6666-6666-666666666666",
            ),
        ),
        # Filter on element and worker run should give the first transcription
        (
            {
                "element": CachedElement(
                    id="11111111-1111-1111-1111-111111111111", type="page"
                ),
                "worker_run": "56785678-5678-5678-5678-567856785678",
            },
            ("11111111-1111-1111-1111-111111111111",),
        ),
        # Filter on element, manual worker run should give the sixth transcription
        (
            {
                "element": CachedElement(
                    id="11111111-1111-1111-1111-111111111111", type="page"
                ),
                "worker_run": False,
            },
            ("66666666-6666-6666-6666-666666666666",),
        ),
        # Filter recursively on element should give all transcriptions inserted
        (
            {
                "element": CachedElement(
                    id="11111111-1111-1111-1111-111111111111", type="page"
                ),
                "recursive": True,
            },
            (
                "11111111-1111-1111-1111-111111111111",
                "22222222-2222-2222-2222-222222222222",
                "33333333-3333-3333-3333-333333333333",
                "44444444-4444-4444-4444-444444444444",
                "55555555-5555-5555-5555-555555555555",
                "66666666-6666-6666-6666-666666666666",
            ),
        ),
        # Filter recursively on element and element_type should give three transcriptions
        (
            {
                "element": CachedElement(
                    id="11111111-1111-1111-1111-111111111111", type="page"
                ),
                "element_type": "something_else",
                "recursive": True,
            },
            (
                "22222222-2222-2222-2222-222222222222",
                "44444444-4444-4444-4444-444444444444",
                "55555555-5555-5555-5555-555555555555",
            ),
        ),
    ],
)
def test_list_transcriptions_with_cache(
    responses, mock_elements_worker_with_cache, filters, expected_ids
):
    # Check we have 5 elements already present in database
    assert CachedTranscription.select().count() == 6

    # Query database through cache
    transcriptions = mock_elements_worker_with_cache.list_transcriptions(**filters)
    assert transcriptions.count() == len(expected_ids)
    for transcription, expected_id in zip(
        transcriptions.order_by(CachedTranscription.id), expected_ids, strict=True
    ):
        assert transcription.id == UUID(expected_id)

    # Check the worker never hits the API for elements
    assert len(responses.calls) == len(BASE_API_CALLS)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS


@pytest.mark.usefixtures("_mock_cached_transcriptions")
@pytest.mark.parametrize(
    ("filters", "expected_ids"),
    [
        # Filter on element and worker_version should give first transcription
        (
            {
                "element": CachedElement(
                    id="11111111-1111-1111-1111-111111111111", type="page"
                ),
                "worker_version": "56785678-5678-5678-5678-567856785678",
            },
            ("11111111-1111-1111-1111-111111111111",),
        ),
        # Filter recursively on element and worker_version should give four transcriptions
        (
            {
                "element": CachedElement(
                    id="11111111-1111-1111-1111-111111111111", type="page"
                ),
                "worker_version": "90129012-9012-9012-9012-901290129012",
                "recursive": True,
            },
            (
                "22222222-2222-2222-2222-222222222222",
                "33333333-3333-3333-3333-333333333333",
                "44444444-4444-4444-4444-444444444444",
                "55555555-5555-5555-5555-555555555555",
            ),
        ),
        # Filter on element with manually created transcription should give sixth transcription
        (
            {
                "element": CachedElement(
                    id="11111111-1111-1111-1111-111111111111", type="page"
                ),
                "worker_version": False,
            },
            ("66666666-6666-6666-6666-666666666666",),
        ),
    ],
)
def test_list_transcriptions_with_cache_deprecation(
    responses, mock_elements_worker_with_cache, filters, expected_ids
):
    # Check we have 5 elements already present in database
    assert CachedTranscription.select().count() == 6

    with pytest.deprecated_call(
        match="`worker_version` usage is deprecated. Consider using `worker_run` instead."
    ):
        # Query database through cache
        transcriptions = mock_elements_worker_with_cache.list_transcriptions(**filters)
    assert transcriptions.count() == len(expected_ids)
    for transcription, expected_id in zip(
        transcriptions.order_by(CachedTranscription.id), expected_ids, strict=True
    ):
        assert transcription.id == UUID(expected_id)

    # Check the worker never hits the API for elements
    assert len(responses.calls) == len(BASE_API_CALLS)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS



# API calls during worker configuration
BASE_API_CALLS = [
    (
        "GET",
        "http://testserver/api/v1/process/workers/56785678-5678-5678-5678-567856785678/",
    ),
    (
        "GET",
        "http://testserver/api/v1/workers/runs/56785678-5678-5678-5678-567856785678/configuration/",
    ),
]



import tempfile
import uuid
from pathlib import Path

import pytest
from requests import HTTPError

from arkindex.exceptions import ErrorResponse
from arkindex_worker.models import Artifact
from tests import FIXTURES_DIR
from tests.test_elements_worker import BASE_API_CALLS

TASK_ID = uuid.UUID("cafecafe-cafe-cafe-cafe-cafecafecafe")


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Task ID
        (
            {"task_id": None},
            "task_id shouldn't be null and should be an UUID",
        ),
        (
            {"task_id": "12341234-1234-1234-1234-123412341234"},
            "task_id shouldn't be null and should be an UUID",
        ),
    ],
)
def test_list_artifacts_wrong_param_task_id(mock_dataset_worker, payload, error):
    with pytest.raises(AssertionError, match=error):
        mock_dataset_worker.list_artifacts(**payload)


def test_list_artifacts_api_error(responses, mock_dataset_worker):
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/task/{TASK_ID}/artifacts/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_dataset_worker.list_artifacts(task_id=TASK_ID)

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("GET", f"http://testserver/api/v1/task/{TASK_ID}/artifacts/")
    ]


def test_list_artifacts(
    responses,
    mock_dataset_worker,
):
    expected_results = [
        {
            "id": "artifact_1",
            "path": "dataset_id.tar.zst",
            "size": 42,
            "content_type": "application/zstd",
            "s3_put_url": None,
            "created": "2000-01-01T00:00:00Z",
            "updated": "2000-01-01T00:00:00Z",
        },
        {
            "id": "artifact_2",
            "path": "logs.log",
            "size": 42,
            "content_type": "text/plain",
            "s3_put_url": None,
            "created": "2000-01-01T00:00:00Z",
            "updated": "2000-01-01T00:00:00Z",
        },
    ]
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/task/{TASK_ID}/artifacts/",
        status=200,
        json=expected_results,
    )

    for idx, artifact in enumerate(mock_dataset_worker.list_artifacts(task_id=TASK_ID)):
        assert isinstance(artifact, Artifact)
        assert artifact == expected_results[idx]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("GET", f"http://testserver/api/v1/task/{TASK_ID}/artifacts/"),
    ]


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Task ID
        (
            {"task_id": None},
            "task_id shouldn't be null and should be an UUID",
        ),
        (
            {"task_id": "12341234-1234-1234-1234-123412341234"},
            "task_id shouldn't be null and should be an UUID",
        ),
    ],
)
def test_download_artifact_wrong_param_task_id(
    mock_dataset_worker, default_artifact, payload, error
):
    api_payload = {
        "task_id": TASK_ID,
        "artifact": default_artifact,
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_dataset_worker.download_artifact(**api_payload)


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Artifact
        (
            {"artifact": None},
            "artifact shouldn't be null and should be an Artifact",
        ),
        (
            {"artifact": "not artifact type"},
            "artifact shouldn't be null and should be an Artifact",
        ),
    ],
)
def test_download_artifact_wrong_param_artifact(
    mock_dataset_worker, default_artifact, payload, error
):
    api_payload = {
        "task_id": TASK_ID,
        "artifact": default_artifact,
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_dataset_worker.download_artifact(**api_payload)


def test_download_artifact_api_error(responses, mock_dataset_worker, default_artifact):
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/task/{TASK_ID}/artifact/dataset_id.tar.zst",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_dataset_worker.download_artifact(
            task_id=TASK_ID, artifact=default_artifact
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("GET", f"http://testserver/api/v1/task/{TASK_ID}/artifact/dataset_id.tar.zst")
    ]


def test_download_artifact(
    responses,
    mock_dataset_worker,
    default_artifact,
):
    archive_path = (
        FIXTURES_DIR
        / "extract_parent_archives"
        / "first_parent"
        / "arkindex_data.tar.zst"
    )
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/task/{TASK_ID}/artifact/dataset_id.tar.zst",
        status=200,
        body=archive_path.read_bytes(),
        content_type="application/zstd",
    )

    assert (
        mock_dataset_worker.download_artifact(
            task_id=TASK_ID, artifact=default_artifact
        ).read()
        == archive_path.read_bytes()
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("GET", f"http://testserver/api/v1/task/{TASK_ID}/artifact/dataset_id.tar.zst"),
    ]


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Path
        (
            {"path": None},
            "path shouldn't be null, should be a Path and should exist",
        ),
        (
            {"path": "not path type"},
            "path shouldn't be null, should be a Path and should exist",
        ),
        (
            {"path": Path("i_do_no_exist.oops")},
            "path shouldn't be null, should be a Path and should exist",
        ),
    ],
)
def test_upload_artifact_wrong_param_path(mock_dataset_worker, payload, error):
    with pytest.raises(AssertionError, match=error):
        mock_dataset_worker.upload_artifact(**payload)


@pytest.fixture
def tmp_file(mock_dataset_worker):
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".txt", dir=mock_dataset_worker.work_dir
    ) as file:
        file.write("Some content...")
        file.seek(0)

        yield Path(file.name)


def test_upload_artifact_api_error(responses, mock_dataset_worker, tmp_file):
    responses.add(
        responses.POST,
        "http://testserver/api/v1/task/my_task/artifacts/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_dataset_worker.upload_artifact(path=tmp_file)

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [("POST", "http://testserver/api/v1/task/my_task/artifacts/")]


def test_upload_artifact_s3_upload_error(
    responses,
    mock_dataset_worker,
    tmp_file,
):
    responses.add(
        responses.POST,
        "http://testserver/api/v1/task/my_task/artifacts/",
        json={
            "id": "11111111-1111-1111-1111-111111111111",
            "path": tmp_file.name,
            "size": 15,
            "content_type": "text/plain",
            "s3_put_url": "http://example.com/oops.txt",
        },
    )
    responses.add(responses.PUT, "http://example.com/oops.txt", status=500)

    with pytest.raises(HTTPError):
        mock_dataset_worker.upload_artifact(path=tmp_file)

    assert len(responses.calls) == len(BASE_API_CALLS) + 2
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("POST", "http://testserver/api/v1/task/my_task/artifacts/"),
        ("PUT", "http://example.com/oops.txt"),
    ]


def test_upload_artifact(
    responses,
    mock_dataset_worker,
    tmp_file,
):
    responses.add(
        responses.POST,
        "http://testserver/api/v1/task/my_task/artifacts/",
        json={
            "id": "11111111-1111-1111-1111-111111111111",
            "path": tmp_file.name,
            "size": 15,
            "content_type": "text/plain",
            "s3_put_url": "http://example.com/test.txt",
        },
    )
    responses.add(responses.PUT, "http://example.com/test.txt")

    mock_dataset_worker.upload_artifact(path=tmp_file)

    assert len(responses.calls) == len(BASE_API_CALLS) + 2
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("POST", "http://testserver/api/v1/task/my_task/artifacts/"),
        ("PUT", "http://example.com/test.txt"),
    ]



import json
import re

import pytest
from responses import matchers

from arkindex.exceptions import ErrorResponse
from arkindex_worker.cache import (
    CachedElement,
    CachedImage,
)
from arkindex_worker.models import Element
from arkindex_worker.worker.element import MissingElementType
from tests import CORPUS_ID

from . import BASE_API_CALLS


def test_list_corpus_types(responses, mock_elements_worker):
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/corpus/{CORPUS_ID}/",
        json={
            "id": CORPUS_ID,
            "types": [{"slug": "folder"}, {"slug": "page"}],
        },
    )

    mock_elements_worker.list_corpus_types()

    assert mock_elements_worker.corpus_types == {
        "folder": {"slug": "folder"},
        "page": {"slug": "page"},
    }


def test_create_element_type_wrong_slug(mock_elements_worker):
    with pytest.raises(
        AssertionError, match="slug shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_element_type(slug=None, name="page")

    with pytest.raises(
        AssertionError, match="slug shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_element_type(slug=1234, name="page")


def test_create_element_type_wrong_name(mock_elements_worker):
    with pytest.raises(
        AssertionError, match="name shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_element_type(slug="page", name=None)

    with pytest.raises(
        AssertionError, match="name shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_element_type(slug="page", name=1234)


def test_create_element_type_wrong_is_folder(mock_elements_worker):
    with pytest.raises(
        AssertionError, match="is_folder shouldn't be null and should be of type bool"
    ):
        mock_elements_worker.create_element_type(
            slug="page", name="page", is_folder=None
        )

    with pytest.raises(
        AssertionError, match="is_folder shouldn't be null and should be of type bool"
    ):
        mock_elements_worker.create_element_type(
            slug="page", name="page", is_folder=1234
        )


def test_create_element_type_api_error(responses, mock_elements_worker):
    responses.add(
        responses.POST,
        "http://testserver/api/v1/elements/type/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_element_type(slug="page", name="page")

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [("POST", "http://testserver/api/v1/elements/type/")]


def test_create_element_type_already_exists(responses, mock_elements_worker):
    assert mock_elements_worker.corpus_types == {}

    responses.add(
        responses.POST,
        "http://testserver/api/v1/elements/type/",
        status=400,
        match=[
            matchers.json_params_matcher(
                {
                    "slug": "page",
                    "display_name": "page",
                    "folder": False,
                    "corpus": CORPUS_ID,
                }
            )
        ],
    )
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/corpus/{CORPUS_ID}/",
        status=200,
        json={
            "id": CORPUS_ID,
            "types": [{"slug": "folder"}, {"slug": "page"}],
        },
    )

    mock_elements_worker.create_element_type(slug="page", name="page")

    assert len(responses.calls) == len(BASE_API_CALLS) + 2
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("POST", "http://testserver/api/v1/elements/type/"),
        ("GET", f"http://testserver/api/v1/corpus/{CORPUS_ID}/"),
    ]

    # Make sure the corpus_types attribute has been updated
    assert mock_elements_worker.corpus_types == {
        "folder": {"slug": "folder"},
        "page": {"slug": "page"},
    }


def test_create_element_type(responses, mock_elements_worker):
    assert mock_elements_worker.corpus_types == {}

    responses.add(
        responses.POST,
        "http://testserver/api/v1/elements/type/",
        status=200,
        match=[
            matchers.json_params_matcher(
                {
                    "slug": "page",
                    "display_name": "page",
                    "folder": False,
                    "corpus": CORPUS_ID,
                }
            )
        ],
        json={"id": "page-id", "slug": "page", "display_name": "page", "folder": False},
    )

    mock_elements_worker.create_element_type(slug="page", name="page")

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("POST", "http://testserver/api/v1/elements/type/"),
    ]

    # Make sure the corpus_types attribute has been updated
    assert mock_elements_worker.corpus_types == {
        "page": {
            "id": "page-id",
            "slug": "page",
            "display_name": "page",
            "folder": False,
        }
    }


def test_check_required_types_wrong_type_slugs(mock_elements_worker):
    with pytest.raises(
        AssertionError, match="type_slugs shouldn't be null and should be of type list"
    ):
        mock_elements_worker.check_required_types(type_slugs=None)

    with pytest.raises(
        AssertionError, match="type_slugs shouldn't be null and should be of type list"
    ):
        mock_elements_worker.check_required_types(type_slugs=1234)

    with pytest.raises(
        AssertionError,
        match="Element type at index 1 in type_slugs: Should be of type str",
    ):
        mock_elements_worker.check_required_types(type_slugs=["page", 1234])


def test_check_required_types_wrong_create_missing(mock_elements_worker):
    with pytest.raises(
        AssertionError,
        match="create_missing shouldn't be null and should be of type bool",
    ):
        mock_elements_worker.check_required_types(
            type_slugs=["page"], create_missing=None
        )

    with pytest.raises(
        AssertionError,
        match="create_missing shouldn't be null and should be of type bool",
    ):
        mock_elements_worker.check_required_types(
            type_slugs=["page"], create_missing=1234
        )


def test_check_required_types_do_not_create_missing(responses, mock_elements_worker):
    # Set one element type
    mock_elements_worker.corpus_types = {"folder": {"slug": "folder"}}

    with pytest.raises(
        MissingElementType, match="Element type `page` was not in the corpus."
    ):
        mock_elements_worker.check_required_types(
            type_slugs=["folder", "page"], create_missing=False
        )

    assert len(responses.calls) == len(BASE_API_CALLS)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS


def test_check_required_types(responses, mock_elements_worker):
    # Set one element type
    mock_elements_worker.corpus_types = {"folder": {"slug": "folder"}}

    # Call to create a new element type
    responses.add(
        responses.POST,
        "http://testserver/api/v1/elements/type/",
        status=200,
        match=[
            matchers.json_params_matcher(
                {
                    "slug": "page",
                    "display_name": "page",
                    "folder": False,
                    "corpus": CORPUS_ID,
                }
            )
        ],
        json={"id": "page-id", "slug": "page", "display_name": "page", "folder": False},
    )

    mock_elements_worker.check_required_types(
        type_slugs=["folder", "page"], create_missing=True
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "POST",
            "http://testserver/api/v1/elements/type/",
        ),
    ]

    # Make sure the element_types attribute has been updated
    assert mock_elements_worker.corpus_types == {
        "folder": {"slug": "folder"},
        "page": {
            "id": "page-id",
            "slug": "page",
            "display_name": "page",
            "folder": False,
        },
    }


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Element
        (
            {"element": None},
            "element shouldn't be null and should be an Element or CachedElement",
        ),
        (
            {"element": "not element type"},
            "element shouldn't be null and should be an Element or CachedElement",
        ),
    ],
)
def test_partial_update_element_wrong_param_element(
    mock_elements_worker, payload, error
):
    api_payload = {
        "element": Element({"zone": None}),
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_elements_worker.partial_update_element(
            **api_payload,
        )


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Type
        ({"type": 1234}, "type should be a str"),
        ({"type": None}, "type should be a str"),
    ],
)
def test_partial_update_element_wrong_param_type(mock_elements_worker, payload, error):
    api_payload = {
        "element": Element({"zone": None}),
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_elements_worker.partial_update_element(
            **api_payload,
        )


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Name
        ({"name": 1234}, "name should be a str"),
        ({"name": None}, "name should be a str"),
    ],
)
def test_partial_update_element_wrong_param_name(mock_elements_worker, payload, error):
    api_payload = {
        "element": Element({"zone": None}),
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_elements_worker.partial_update_element(
            **api_payload,
        )


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Polygon
        ({"polygon": "not a polygon"}, "polygon should be a list"),
        ({"polygon": None}, "polygon should be a list"),
        ({"polygon": [[1, 1], [2, 2]]}, "polygon should have at least three points"),
        (
            {"polygon": [[1, 1, 1], [2, 2, 1], [2, 1, 1], [1, 2, 1]]},
            "polygon points should be lists of two items",
        ),
        (
            {"polygon": [[1], [2], [2], [1]]},
            "polygon points should be lists of two items",
        ),
        (
            {"polygon": [["not a coord", 1], [2, 2], [2, 1], [1, 2]]},
            "polygon points should be lists of two numbers",
        ),
    ],
)
def test_partial_update_element_wrong_param_polygon(
    mock_elements_worker, payload, error
):
    api_payload = {
        "element": Element({"zone": None}),
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_elements_worker.partial_update_element(
            **api_payload,
        )


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Confidence
        ({"confidence": "lol"}, "confidence should be None or a float in [0..1] range"),
        ({"confidence": "0.2"}, "confidence should be None or a float in [0..1] range"),
        ({"confidence": -1.0}, "confidence should be None or a float in [0..1] range"),
        ({"confidence": 1.42}, "confidence should be None or a float in [0..1] range"),
        (
            {"confidence": float("inf")},
            "confidence should be None or a float in [0..1] range",
        ),
    ],
)
def test_partial_update_element_wrong_param_conf(mock_elements_worker, payload, error):
    api_payload = {
        "element": Element({"zone": None}),
        **payload,
    }

    with pytest.raises(AssertionError, match=re.escape(error)):
        mock_elements_worker.partial_update_element(
            **api_payload,
        )


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Rotation angle
        ({"rotation_angle": "lol"}, "rotation_angle should be a positive integer"),
        ({"rotation_angle": -1}, "rotation_angle should be a positive integer"),
        ({"rotation_angle": 0.5}, "rotation_angle should be a positive integer"),
        ({"rotation_angle": None}, "rotation_angle should be a positive integer"),
    ],
)
def test_partial_update_element_wrong_param_rota(mock_elements_worker, payload, error):
    api_payload = {
        "element": Element({"zone": None}),
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_elements_worker.partial_update_element(
            **api_payload,
        )


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Mirrored
        ({"mirrored": "lol"}, "mirrored should be a boolean"),
        ({"mirrored": 1234}, "mirrored should be a boolean"),
        ({"mirrored": None}, "mirrored should be a boolean"),
    ],
)
def test_partial_update_element_wrong_param_mir(mock_elements_worker, payload, error):
    api_payload = {
        "element": Element({"zone": None}),
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_elements_worker.partial_update_element(
            **api_payload,
        )


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Image
        ({"image": "lol"}, "image should be a UUID"),
        ({"image": 1234}, "image should be a UUID"),
        ({"image": None}, "image should be a UUID"),
    ],
)
def test_partial_update_element_wrong_param_image(mock_elements_worker, payload, error):
    api_payload = {
        "element": Element({"zone": None}),
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_elements_worker.partial_update_element(
            **api_payload,
        )


def test_partial_update_element_api_error(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.PATCH,
        f"http://testserver/api/v1/element/{elt.id}/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.partial_update_element(
            element=elt,
            type="something",
            name="0",
            polygon=[[1, 1], [2, 2], [2, 1], [1, 2]],
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [("PATCH", f"http://testserver/api/v1/element/{elt.id}/")]


@pytest.mark.usefixtures("_mock_cached_elements", "_mock_cached_images")
@pytest.mark.parametrize(
    "payload",
    [
        (
            {
                "polygon": [[10, 10], [20, 20], [20, 10], [10, 20]],
                "confidence": None,
            }
        ),
        (
            {
                "rotation_angle": 45,
                "mirrored": False,
            }
        ),
        (
            {
                "polygon": [[10, 10], [20, 20], [20, 10], [10, 20]],
                "confidence": None,
                "rotation_angle": 45,
                "mirrored": False,
            }
        ),
    ],
)
def test_partial_update_element(responses, mock_elements_worker_with_cache, payload):
    elt = CachedElement.select().first()
    new_image = CachedImage.select().first()

    elt_response = {
        "image": str(new_image.id),
        **payload,
    }
    responses.add(
        responses.PATCH,
        f"http://testserver/api/v1/element/{elt.id}/",
        status=200,
        # UUID not allowed in JSON
        json=elt_response,
    )

    element_update_response = mock_elements_worker_with_cache.partial_update_element(
        element=elt,
        **{**elt_response, "image": new_image.id},
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "PATCH",
            f"http://testserver/api/v1/element/{elt.id}/",
        ),
    ]
    assert json.loads(responses.calls[-1].request.body) == elt_response
    assert element_update_response == elt_response

    cached_element = CachedElement.get(CachedElement.id == elt.id)
    # Always present in payload
    assert str(cached_element.image_id) == elt_response["image"]
    # Optional params
    if "polygon" in payload:
        # Cast to string as this is the only difference compared to model
        elt_response["polygon"] = str(elt_response["polygon"])

    for param in payload:
        assert getattr(cached_element, param) == elt_response[param]


@pytest.mark.usefixtures("_mock_cached_elements")
@pytest.mark.parametrize("confidence", [None, 0.42])
def test_partial_update_element_confidence(
    responses, mock_elements_worker_with_cache, confidence
):
    elt = CachedElement.select().first()
    elt_response = {
        "polygon": [[10, 10], [20, 20], [20, 10], [10, 20]],
        "confidence": confidence,
    }
    responses.add(
        responses.PATCH,
        f"http://testserver/api/v1/element/{elt.id}/",
        status=200,
        json=elt_response,
    )

    element_update_response = mock_elements_worker_with_cache.partial_update_element(
        element=elt,
        **elt_response,
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "PATCH",
            f"http://testserver/api/v1/element/{elt.id}/",
        ),
    ]
    assert json.loads(responses.calls[-1].request.body) == elt_response
    assert element_update_response == elt_response

    cached_element = CachedElement.get(CachedElement.id == elt.id)
    assert cached_element.polygon == str(elt_response["polygon"])
    assert cached_element.confidence == confidence



"""
BaseWorker methods for datasets.
"""

import uuid
from argparse import ArgumentTypeError
from collections.abc import Iterator
from enum import Enum

from arkindex_worker import logger
from arkindex_worker.cache import unsupported_cache
from arkindex_worker.models import Dataset, Element, Set


class DatasetState(Enum):
    """
    State of a dataset.
    """

    Open = "open"
    """
    The dataset is open.
    """

    Building = "building"
    """
    The dataset is being built.
    """

    Complete = "complete"
    """
    The dataset is complete.
    """

    Error = "error"
    """
    The dataset is in error.
    """


class MissingDatasetArchive(Exception):
    """
    Exception raised when the compressed archive associated to
    a dataset isn't found in its task artifacts.
    """


def check_dataset_set(value: str) -> tuple[uuid.UUID, str]:
    """The `--set` argument should have the following format:
    <dataset_id>:<set_name>

    Args:
        value (str): Provided argument.

    Raises:
        ArgumentTypeError: When the value is invalid.

    Returns:
        tuple[uuid.UUID, str]: The ID of the dataset parsed as UUID and the name of the set.
    """
    values = value.split(":")
    if len(values) != 2:
        raise ArgumentTypeError(
            f"'{value}' is not in the correct format `<dataset_id>:<set_name>`"
        )

    dataset_id, set_name = values
    try:
        dataset_id = uuid.UUID(dataset_id)
        return (dataset_id, set_name)
    except (TypeError, ValueError) as e:
        raise ArgumentTypeError(f"'{dataset_id}' should be a valid UUID") from e


class DatasetMixin:
    def add_arguments(self) -> None:
        """Define specific ``argparse`` arguments for the worker using this mixin"""
        self.parser.add_argument(
            "--set",
            type=check_dataset_set,
            nargs="+",
            help="""
                One or more Arkindex dataset sets, format is <dataset_uuid>:<set_name>
                (e.g.: "12341234-1234-1234-1234-123412341234:train")
            """,
            default=[],
        )
        super().add_arguments()

    def list_process_sets(self) -> Iterator[Set]:
        """
        List dataset sets associated to the worker's process. This helper is not available in developer mode.

        :returns: An iterator of ``Set`` objects built from the ``ListProcessSets`` API endpoint.
        """
        assert not self.is_read_only, "This helper is not available in read-only mode."

        results = self.api_client.paginate(
            "ListProcessSets", id=self.process_information["id"]
        )

        return map(
            lambda result: Set(
                name=result["set_name"], dataset=Dataset(**result["dataset"])
            ),
            results,
        )

    def list_set_elements(self, dataset_set: Set) -> Iterator[Element]:
        """
        List elements in a dataset set.

        :param dataset_set: Set to find elements in.
        :returns: An iterator of Element built from the ``ListDatasetElements`` API endpoint.
        """
        assert dataset_set and isinstance(dataset_set, Set), (
            "dataset_set shouldn't be null and should be a Set"
        )

        results = self.api_client.paginate(
            "ListDatasetElements", id=dataset_set.dataset.id, set=dataset_set.name
        )

        return map(lambda result: Element(**result["element"]), results)

    def list_sets(self) -> Iterator[Set]:
        """
        List the sets to be processed, either from the CLI arguments or using the
        [list_process_sets][arkindex_worker.worker.dataset.DatasetMixin.list_process_sets] method.

        :returns: An iterator of ``Set`` objects.
        """
        if not self.is_read_only:
            yield from self.list_process_sets()

        datasets: dict[uuid.UUID, Dataset] = {}
        for dataset_id, set_name in self.args.set:
            # Retrieving dataset information if not already cached
            if dataset_id not in datasets:
                datasets[dataset_id] = Dataset(
                    **self.api_client.request("RetrieveDataset", id=dataset_id)
                )

            yield Set(name=set_name, dataset=datasets[dataset_id])

    @unsupported_cache
    def update_dataset_state(self, dataset: Dataset, state: DatasetState) -> Dataset:
        """
        Partially updates a dataset state through the API.

        :param dataset: The dataset to update.
        :param state: State of the dataset.
        :returns: The updated ``Dataset`` object from the ``PartialUpdateDataset`` API endpoint.
        """
        assert dataset and isinstance(dataset, Dataset), (
            "dataset shouldn't be null and should be a Dataset"
        )
        assert state and isinstance(state, DatasetState), (
            "state shouldn't be null and should be a str from DatasetState"
        )

        if self.is_read_only:
            logger.warning("Cannot update dataset as this worker is in read-only mode")
            return

        updated_dataset = self.api_client.request(
            "PartialUpdateDataset",
            id=dataset.id,
            body={"state": state.value},
        )
        dataset.update(updated_dataset)

        return dataset
